@@ -1,7 +1,7 @@
 /*
  * Copyright (C) 2010-2011 Robert Ancell.
  * Author: Robert Ancell <robert.ancell@canonical.com>
- * 
+ *
  * This program is free software: you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free Software
  * Foundation, either version 3 of the License, or (at your option) any later
@@ -39,9 +39,9 @@ xephyr_setup_cb (gpointer user_data)
 static void
 xephyr_signal_cb (int signum)
 {
-    gchar *path;
-    GVariant *result;
-    GError *error = NULL;
+    g_autofree gchar *path = NULL;
+    g_autoptr(GVariant) result = NULL;
+    g_autoptr(GError) error = NULL;
 
     result = g_dbus_proxy_call_sync (dm_proxy,
                                      "AddLocalXSeat",
@@ -72,11 +72,11 @@ xephyr_signal_cb (int signum)
 static GDBusProxy *
 get_seat_proxy (void)
 {
-    GError *error = NULL;
+    g_autoptr(GError) error = NULL;
 
     if (seat_proxy)
         return seat_proxy;
-  
+
     if (!g_getenv ("XDG_SEAT_PATH"))
     {
         g_printerr ("Not running inside a display manager, XDG_SEAT_PATH not defined\n");
@@ -96,8 +96,7 @@ get_seat_proxy (void)
         g_printerr ("Unable to contact display manager: %s\n", error->message);
         exit (EXIT_FAILURE);
     }
-    g_clear_error (&error);
-  
+
     return seat_proxy;
 }
 
@@ -107,7 +106,7 @@ main (int argc, char **argv)
     gchar *command;
     gint n_options;
     gchar **options;
-    GError *error = NULL;
+    g_autoptr(GError) error = NULL;
     gint arg_index;
 
 #if !defined(GLIB_VERSION_2_36)
@@ -120,7 +119,7 @@ main (int argc, char **argv)
 
         if (!g_str_has_prefix (arg, "-"))
             break;
-      
+
         if (strcmp (arg, "-h") == 0 || strcmp (arg, "--help") == 0)
         {
             g_printerr ("Usage:\n"
@@ -178,8 +177,7 @@ main (int argc, char **argv)
         g_printerr ("Unable to contact display manager: %s\n", error->message);
         return EXIT_FAILURE;
     }
-    g_clear_error (&error);
-  
+
     command = argv[arg_index];
     arg_index++;
     n_options = argc - arg_index;
@@ -300,7 +298,7 @@ main (int argc, char **argv)
         while (g_variant_iter_loop (seat_iter, "&o", &seat_path))
         {
             gchar *seat_name;
-            GDBusProxy *proxy;
+            g_autoptr(GDBusProxy) proxy = NULL;
             gchar **property_names;
             GVariant *sessions;
             GVariantIter *session_iter;
@@ -327,14 +325,13 @@ main (int argc, char **argv)
             property_names = g_dbus_proxy_get_cached_property_names (proxy);
             for (i = 0; property_names[i]; i++)
             {
-                GVariant *value;
+                g_autoptr(GVariant) value = NULL;
 
                 if (strcmp (property_names[i], "Sessions") == 0)
                     continue;
 
                 value = g_dbus_proxy_get_cached_property (proxy, property_names[i]);
                 g_print ("  %s=%s\n", property_names[i], g_variant_print (value, FALSE));
-                g_variant_unref (value);
             }
 
             sessions = g_dbus_proxy_get_cached_property (proxy, "Sessions");
@@ -344,8 +341,8 @@ main (int argc, char **argv)
             g_variant_get (sessions, "ao", &session_iter);
             while (g_variant_iter_loop (session_iter, "&o", &session_path))
             {
-                GDBusProxy *session_proxy;
-                gchar *session_name;
+                g_autoptr(GDBusProxy) session_proxy = NULL;
+                const gchar *session_name;
 
                 if (g_str_has_prefix (session_path, "/org/freedesktop/DisplayManager/"))
                     session_name = session_path + strlen ("/org/freedesktop/DisplayManager/");
@@ -367,21 +364,16 @@ main (int argc, char **argv)
                 property_names = g_dbus_proxy_get_cached_property_names (session_proxy);
                 for (i = 0; property_names[i]; i++)
                 {
-                    GVariant *value;
+                    g_autoptr(GVariant) value = NULL;
 
                     if (strcmp (property_names[i], "Seat") == 0)
                         continue;
 
                     value = g_dbus_proxy_get_cached_property (session_proxy, property_names[i]);
                     g_print ("    %s=%s\n", property_names[i], g_variant_print (value, FALSE));
-                    g_variant_unref (value);
                 }
-
-                g_object_unref (session_proxy);
             }
             g_variant_iter_free (session_iter);
-
-            g_object_unref (proxy);
         }
         g_variant_iter_free (seat_iter);
 
@@ -423,13 +415,12 @@ main (int argc, char **argv)
         xephyr_display_number = 0;
         while (TRUE)
         {
-            gchar *lock_name;
+            g_autofree gchar *lock_name = NULL;
             gboolean has_lock;
 
             lock_name = g_strdup_printf ("/tmp/.X%d-lock", xephyr_display_number);
             has_lock = g_file_test (lock_name, G_FILE_TEST_EXISTS);
-            g_free (lock_name);
-          
+
             if (has_lock)
                 xephyr_display_number++;
             else
@@ -445,11 +436,11 @@ main (int argc, char **argv)
         }
         else if (strcmp (dimensions, "fullscreen") == 0)
         {
-            xephyr_command = g_strdup_printf ("Xephyr :%d -fullscreen", xephyr_display_number);   
+            xephyr_command = g_strdup_printf ("Xephyr :%d -fullscreen", xephyr_display_number);
         }
         else
         {
-            xephyr_command = g_strdup_printf ("Xephyr :%d -screen %s", xephyr_display_number, dimensions); 
+            xephyr_command = g_strdup_printf ("Xephyr :%d -screen %s", xephyr_display_number, dimensions);
         }
         if (!g_shell_parse_argv (xephyr_command, NULL, &xephyr_argv, &error) ||
             !g_spawn_async (NULL, xephyr_argv, NULL,
@@ -460,7 +451,6 @@ main (int argc, char **argv)
             g_printerr ("Error running Xephyr: %s\n", error->message);
             exit (EXIT_FAILURE);
         }
-        g_clear_error (&error);
 
         /* Block until ready */
         loop = g_main_loop_new (NULL, FALSE);
@@ -503,7 +493,7 @@ main (int argc, char **argv)
         g_variant_get (result, "(&o)", &path);
         g_print ("%s\n", path);
 
-        return EXIT_SUCCESS; 
+        return EXIT_SUCCESS;
     }
     else if (strcmp (command, "add-seat") == 0)
     {
@@ -521,10 +511,11 @@ main (int argc, char **argv)
 
         type = options[0];
         properties = g_variant_builder_new (G_VARIANT_TYPE ("a(ss)"));
-      
+
         for (i = 1; i < n_options; i++)
         {
-            gchar *property, *name, *value;
+            g_autofree gchar *property = NULL;
+            gchar *name, *value;
 
             property = g_strdup (options[i]);
             name = property;
@@ -538,9 +529,8 @@ main (int argc, char **argv)
                value = "";
 
             g_variant_builder_add_value (properties, g_variant_new ("(ss)", name, value));
-            g_free (property);
         }
-      
+
         result = g_dbus_proxy_call_sync (dm_proxy,
                                          "AddSeat",
                                          g_variant_new ("(sa(ss))", type, properties),
