@@ -17,11 +17,16 @@
 #include <utmpx.h>
 #include <sys/mman.h>
 
+#if HAVE_LIBAUDIT
+#include <libaudit.h>
+#endif
+
 #include "configuration.h"
 #include "session-child.h"
 #include "session.h"
 #include "console-kit.h"
 #include "login1.h"
+#include "log-file.h"
 #include "privileges.h"
 #include "x-authority.h"
 #include "configuration.h"
@@ -67,7 +72,7 @@ read_data (void *buf, size_t count)
     n_read = read (from_daemon_output, buf, count);
     if (n_read < 0)
         g_printerr ("Error reading from daemon: %s\n", strerror (errno));
-  
+
     return n_read;
 }
 
@@ -86,10 +91,10 @@ read_string_full (void* (*alloc_fn)(size_t n))
         g_printerr ("Invalid string length %d from daemon\n", length);
         return NULL;
     }
-  
+
     value = (*alloc_fn) (sizeof (char) * (length + 1));
     read_data (value, length);
-    value[length] = '\0';      
+    value[length] = '\0';
 
     return value;
 }
@@ -216,10 +221,36 @@ updwtmpx (const gchar *wtmp_file, struct utmpx *ut)
         strncpy (u.ut_host, ut->ut_host, sizeof (u.ut_host));
     u.ut_tv.tv_sec = ut->ut_tv.tv_sec;
     u.ut_tv.tv_usec = ut->ut_tv.tv_usec;
-  
+
     updwtmp (wtmp_file, &u);
 }
 
+#if HAVE_LIBAUDIT
+static void
+audit_event (int type, const gchar *username, uid_t uid, const gchar *remote_host_name, const gchar *tty, gboolean success)
+{
+    int auditfd, result;
+    const char *op = NULL;
+
+    auditfd = audit_open ();
+    if (auditfd < 0) {
+        g_printerr ("Error opening audit socket: %s\n", strerror (errno));
+        return;
+    }
+
+    if (type == AUDIT_USER_LOGIN)
+        op = "login";
+    else if (type == AUDIT_USER_LOGOUT)
+        op = "logout";
+    result = success == TRUE ? 1 : 0;
+
+    if (audit_log_acct_message (auditfd, type, NULL, op, username, uid, remote_host_name, NULL, tty, result) <= 0)
+        g_printerr ("Error writing audit message: %s\n", strerror (errno));
+
+    close (auditfd);
+}
+#endif
+
 int
 session_child_run (int argc, char **argv)
 {
@@ -227,7 +258,8 @@ session_child_run (int argc, char **argv)
     int i, version, fd, result;
     gboolean auth_complete = TRUE;
     User *user = NULL;
-    gchar *log_filename, *log_backup_filename = NULL;
+    g_autofree gchar *log_filename = NULL;
+    LogMode log_mode = LOG_MODE_BACKUP_AND_TRUNCATE;
     gsize env_length;
     gsize command_argc;
     gchar **command_argv;
@@ -243,11 +275,16 @@ session_child_run (int argc, char **argv)
     XAuthority *x_authority = NULL;
     gchar *x_authority_filename;
     GDBusConnection *bus;
-    const gchar *login1_session = NULL;
+    const gchar *login1_session_id = NULL;
     gchar *console_kit_cookie = NULL;
     const gchar *locale_value;
-    gchar *locale_var;
     static const gchar * const locale_var_names[] = {
+        "LC_PAPER",
+        "LC_NAME",
+        "LC_ADDRESS",
+        "LC_TELEPHONE",
+        "LC_MEASUREMENT",
+        "LC_IDENTIFICATION",
         "LC_COLLATE",
         "LC_CTYPE",
         "LC_MONETARY",
@@ -261,7 +298,7 @@ session_child_run (int argc, char **argv)
     gid_t gid;
     uid_t uid;
     const gchar *home_directory;
-    GError *error = NULL;
+    g_autoptr(GError) error = NULL;
 
 #if !defined(GLIB_VERSION_2_36)
     g_type_init ();
@@ -329,7 +366,7 @@ session_child_run (int argc, char **argv)
         pam_set_item (pam_handle, PAM_TTY, xdisplay);
     }
     else if (tty)
-        pam_set_item (pam_handle, PAM_TTY, tty);    
+        pam_set_item (pam_handle, PAM_TTY, tty);
 
 #ifdef PAM_XAUTHDATA
     if (x_authority)
@@ -370,11 +407,8 @@ session_child_run (int argc, char **argv)
             ut.ut_type = USER_PROCESS;
             ut.ut_pid = getpid ();
             if (xdisplay)
-            {
-                strncpy (ut.ut_line, xdisplay, sizeof (ut.ut_line));
                 strncpy (ut.ut_id, xdisplay, sizeof (ut.ut_id));
-            }
-            else if (tty)
+            if (tty && g_str_has_prefix (tty, "/dev/"))
                 strncpy (ut.ut_line, tty + strlen ("/dev/"), sizeof (ut.ut_line));
             strncpy (ut.ut_user, username, sizeof (ut.ut_user));
             if (xdisplay)
@@ -386,6 +420,10 @@ session_child_run (int argc, char **argv)
             ut.ut_tv.tv_usec = tv.tv_usec;
 
             updwtmpx ("/var/log/btmp", &ut);
+
+#if HAVE_LIBAUDIT
+            audit_event (AUDIT_USER_LOGIN, username, -1, remote_host_name, tty, FALSE);
+#endif
         }
 
         /* Check account is valid */
@@ -421,9 +459,8 @@ session_child_run (int argc, char **argv)
             {
                 if ((locale_value = g_getenv (locale_var_names[i])) != NULL)
                 {
-                    locale_var = g_strdup_printf ("%s=%s", locale_var_names[i], locale_value);
+                    g_autofree gchar *locale_var = g_strdup_printf ("%s=%s", locale_var_names[i], locale_value);
                     pam_putenv (pam_handle, locale_var);
-                    g_free (locale_var);
                 }
             }
         }
@@ -454,18 +491,19 @@ session_child_run (int argc, char **argv)
 
     /* Get the command to run (blocks) */
     log_filename = read_string ();
+    if (version >= 3)
+        read_data (&log_mode, sizeof (log_mode));
     if (version >= 1)
     {
         g_free (tty);
-        tty = read_string ();      
+        tty = read_string ();
     }
     x_authority_filename = read_string ();
     if (version >= 1)
     {
         g_free (xdisplay);
         xdisplay = read_string ();
-        if (x_authority)
-            g_object_unref (x_authority);
+        g_clear_object (&x_authority);
         x_authority = read_xauth ();
     }
     read_data (&env_length, sizeof (env_length));
@@ -488,20 +526,17 @@ session_child_run (int argc, char **argv)
     /* Redirect stderr to a log file */
     if (log_filename)
     {
-        log_backup_filename = g_strdup_printf ("%s.old", log_filename);
         if (g_path_is_absolute (log_filename))
         {
-            rename (log_filename, log_backup_filename);
-            fd = open (log_filename, O_WRONLY | O_APPEND | O_CREAT, 0600);
+            fd = log_file_open (log_filename, log_mode);
             dup2 (fd, STDERR_FILENO);
             close (fd);
-            g_free (log_filename);
-            log_filename = NULL;
+            g_clear_pointer (&log_filename, g_free);
         }
     }
     else
     {
-        fd = open ("/dev/null", O_WRONLY);   
+        fd = open ("/dev/null", O_WRONLY);
         dup2 (fd, STDERR_FILENO);
         close (fd);
     }
@@ -524,7 +559,7 @@ session_child_run (int argc, char **argv)
         pam_end (pam_handle, 0);
         return EXIT_FAILURE;
     }
-     
+
     /* Open the session */
     result = pam_open_session (pam_handle, 0);
     if (result != PAM_SUCCESS)
@@ -544,16 +579,16 @@ session_child_run (int argc, char **argv)
         return EXIT_FAILURE;
     }
 
-    login1_session = pam_getenv (pam_handle, "XDG_SESSION_ID");
-    if (login1_session)
+    /* Check what logind session we are, or fallback to ConsoleKit */
+    login1_session_id = pam_getenv (pam_handle, "XDG_SESSION_ID");
+    if (login1_session_id)
     {
-        write_string (login1_session);
+        write_string (login1_session_id);
         if (version >= 2)
             write_string (NULL);
     }
     else
     {
-        /* Open a Console Kit session */
         g_variant_builder_init (&ck_parameters, G_VARIANT_TYPE ("(a(sv))"));
         g_variant_builder_open (&ck_parameters, G_VARIANT_TYPE ("a(sv)"));
         g_variant_builder_add (&ck_parameters, "(sv)", "unix-user", g_variant_new_int32 (user_get_uid (user)));
@@ -578,10 +613,17 @@ session_child_run (int argc, char **argv)
         write_string (console_kit_cookie);
         if (console_kit_cookie)
         {
-            gchar *value;
+            g_autofree gchar *value = NULL;
+            g_autofree gchar *runtime_dir = NULL;
             value = g_strdup_printf ("XDG_SESSION_COOKIE=%s", console_kit_cookie);
             pam_putenv (pam_handle, value);
-            g_free (value);
+
+            runtime_dir = ck_get_xdg_runtime_dir (console_kit_cookie);
+            if (runtime_dir)
+            {
+                g_autofree gchar *v = g_strdup_printf ("XDG_RUNTIME_DIR=%s", runtime_dir);
+                pam_putenv (pam_handle, v);
+            }
         }
     }
 
@@ -589,8 +631,8 @@ session_child_run (int argc, char **argv)
     if (x_authority)
     {
         gboolean drop_privileges, result;
-        gchar *value;
-        GError *error = NULL;
+        g_autofree gchar *value = NULL;
+        g_autoptr(GError) error = NULL;
 
         drop_privileges = geteuid () == 0;
         if (drop_privileges)
@@ -601,7 +643,6 @@ session_child_run (int argc, char **argv)
 
         if (error)
             g_printerr ("Error writing X authority: %s\n", error->message);
-        g_clear_error (&error);
         if (!result)
         {
             pam_end (pam_handle, 0);
@@ -610,7 +651,6 @@ session_child_run (int argc, char **argv)
 
         value = g_strdup_printf ("XAUTHORITY=%s", x_authority_filename);
         pam_putenv (pam_handle, value);
-        g_free (value);
     }
 
     /* Catch terminate signal and pass it to the child */
@@ -646,8 +686,7 @@ session_child_run (int argc, char **argv)
 
         if (log_filename)
         {
-            rename (log_filename, log_backup_filename);
-            fd = open (log_filename, O_WRONLY | O_APPEND | O_CREAT, 0600);
+            fd = log_file_open (log_filename, log_mode);
             if (fd >= 0)
             {
                 dup2 (fd, STDERR_FILENO);
@@ -655,6 +694,9 @@ session_child_run (int argc, char **argv)
             }
         }
 
+        /* Reset SIGPIPE handler so the child has default behaviour (we disabled it at LightDM start) */
+        signal (SIGPIPE, SIG_DFL);
+
         /* Run the command */
         execve (command_argv[0], command_argv, pam_getenvlist (pam_handle));
         _exit (EXIT_FAILURE);
@@ -670,6 +712,8 @@ session_child_run (int argc, char **argv)
     /* Wait for the command to complete (blocks) */
     if (child_pid > 0)
     {
+        int child_status;
+
         /* Log to utmp */
         if (g_strcmp0 (pam_getenv (pam_handle, "XDG_SESSION_CLASS"), "greeter") != 0)
         {
@@ -680,11 +724,8 @@ session_child_run (int argc, char **argv)
             ut.ut_type = USER_PROCESS;
             ut.ut_pid = child_pid;
             if (xdisplay)
-            {
-                strncpy (ut.ut_line, xdisplay, sizeof (ut.ut_line));
                 strncpy (ut.ut_id, xdisplay, sizeof (ut.ut_id));
-            }
-            else if (tty)
+            if (tty && g_str_has_prefix (tty, "/dev/"))
                 strncpy (ut.ut_line, tty + strlen ("/dev/"), sizeof (ut.ut_line));
             strncpy (ut.ut_user, username, sizeof (ut.ut_user));
             if (xdisplay)
@@ -701,10 +742,18 @@ session_child_run (int argc, char **argv)
                 g_printerr ("Failed to write utmpx: %s\n", strerror (errno));
             endutxent ();
             updwtmpx ("/var/log/wtmp", &ut);
+
+#if HAVE_LIBAUDIT          
+            audit_event (AUDIT_USER_LOGIN, username, uid, remote_host_name, tty, TRUE);
+#endif
         }
 
-        waitpid (child_pid, &return_code, 0);
+        waitpid (child_pid, &child_status, 0);
         child_pid = 0;
+        if (WIFEXITED (child_status))
+            return_code = WEXITSTATUS (child_status);
+        else
+            return_code = EXIT_FAILURE;
 
         /* Log to utmp */
         if (g_strcmp0 (pam_getenv (pam_handle, "XDG_SESSION_CLASS"), "greeter") != 0)
@@ -716,11 +765,8 @@ session_child_run (int argc, char **argv)
             ut.ut_type = DEAD_PROCESS;
             ut.ut_pid = child_pid;
             if (xdisplay)
-            {
-                strncpy (ut.ut_line, xdisplay, sizeof (ut.ut_line));
                 strncpy (ut.ut_id, xdisplay, sizeof (ut.ut_id));
-            }
-            else if (tty)
+            if (tty && g_str_has_prefix (tty, "/dev/"))
                 strncpy (ut.ut_line, tty + strlen ("/dev/"), sizeof (ut.ut_line));
             strncpy (ut.ut_user, username, sizeof (ut.ut_user));
             if (xdisplay)
@@ -737,6 +783,10 @@ session_child_run (int argc, char **argv)
                 g_printerr ("Failed to write utmpx: %s\n", strerror (errno));
             endutxent ();
             updwtmpx ("/var/log/wtmp", &ut);
+
+#if HAVE_LIBAUDIT
+            audit_event (AUDIT_USER_LOGOUT, username, uid, remote_host_name, tty, TRUE);
+#endif
         }
     }
 
@@ -744,7 +794,7 @@ session_child_run (int argc, char **argv)
     if (x_authority)
     {
         gboolean drop_privileges, result;
-        GError *error = NULL;
+        g_autoptr(GError) error = NULL;
 
         drop_privileges = geteuid () == 0;
         if (drop_privileges)
@@ -755,7 +805,6 @@ session_child_run (int argc, char **argv)
 
         if (error)
             g_printerr ("Error removing X authority: %s\n", error->message);
-        g_clear_error (&error);
         if (!result)
             _exit (EXIT_FAILURE);
     }
@@ -768,7 +817,7 @@ session_child_run (int argc, char **argv)
     pam_close_session (pam_handle, 0);
 
     /* Remove credentials */
-    result = pam_setcred (pam_handle, PAM_DELETE_CRED);
+    pam_setcred (pam_handle, PAM_DELETE_CRED);
 
     pam_end (pam_handle, 0);
     pam_handle = NULL;
