@@ -27,8 +27,10 @@
 #include "login1.h"
 #include "guest-account.h"
 #include "shared-data-manager.h"
+#include "greeter-socket.h"
 
 enum {
+    CREATE_GREETER,
     GOT_MESSAGES,
     AUTHENTICATION_COMPLETE,
     STOPPED,
@@ -84,6 +86,7 @@ struct SessionPrivate
 
     /* File to log to */
     gchar *log_filename;
+    LogMode log_mode;
 
     /* tty this session is running on */
     gchar *tty;
@@ -93,14 +96,17 @@ struct SessionPrivate
     XAuthority *x_authority;
     gboolean x_authority_use_system_location;
 
+    /* Socket to allow greeters to connect to (if allowed) */
+    GreeterSocket *greeter_socket;
+
     /* Remote host this session is being controlled from */
     gchar *remote_host_name;
 
     /* Console kit cookie */
     gchar *console_kit_cookie;
 
-    /* login1 session */
-    gchar *login1_session;
+    /* login1 session ID */
+    gchar *login1_session_id;
 
     /* Environment to set in child */
     GList *env;
@@ -135,8 +141,7 @@ session_set_config (Session *session, SessionConfig *config)
 {
     g_return_if_fail (session != NULL);
 
-    if (session->priv->config)
-        g_object_unref (session->priv->config);
+    g_clear_object (&session->priv->config);
     session->priv->config = g_object_ref (config);
 }
 
@@ -151,7 +156,7 @@ const gchar *
 session_get_session_type (Session *session)
 {
     g_return_val_if_fail (session != NULL, NULL);
-    return session_config_get_session_type (session_get_config (session));
+    return session_config_get_session_type (session->priv->config);
 }
 
 void
@@ -199,11 +204,12 @@ session_get_is_guest (Session *session)
 }
 
 void
-session_set_log_file (Session *session, const gchar *filename)
+session_set_log_file (Session *session, const gchar *filename, LogMode log_mode)
 {
     g_return_if_fail (session != NULL);
     g_free (session->priv->log_filename);
     session->priv->log_filename = g_strdup (filename);
+    session->priv->log_mode = log_mode;
 }
 
 void
@@ -211,11 +217,13 @@ session_set_display_server (Session *session, DisplayServer *display_server)
 {
     g_return_if_fail (session != NULL);
     g_return_if_fail (display_server != NULL);
+
+    if (session->priv->display_server == display_server)
+        return;
+
     if (session->priv->display_server)
-    {
         display_server_disconnect_session (session->priv->display_server, session);
-        g_object_unref (session->priv->display_server);
-    }
+    g_clear_object (&session->priv->display_server);
     session->priv->display_server = g_object_ref (display_server);
 }
 
@@ -246,11 +254,7 @@ void
 session_set_x_authority (Session *session, XAuthority *authority, gboolean use_system_location)
 {
     g_return_if_fail (session != NULL);
-    if (session->priv->x_authority)
-    {
-        g_object_unref (session->priv->x_authority);
-        session->priv->x_authority = NULL;
-    }
+    g_clear_object (&session->priv->x_authority);
     if (authority)
         session->priv->x_authority = g_object_ref (authority);
     session->priv->x_authority_use_system_location = use_system_location;
@@ -310,7 +314,7 @@ session_get_env (Session *session, const gchar *name)
     link = find_env_entry (session, name);
     if (!link)
         return NULL;
-  
+
     entry = link->data;
 
     return entry + strlen (name) + 1;
@@ -322,7 +326,7 @@ session_unset_env (Session *session, const gchar *name)
     GList *link;
 
     g_return_if_fail (session != NULL);
-  
+
     link = find_env_entry (session, name);
     if (!link)
         return;
@@ -467,7 +471,7 @@ static gboolean
 from_child_cb (GIOChannel *source, GIOCondition condition, gpointer data)
 {
     Session *session = data;
-    gchar *username;
+    g_autofree gchar *username = NULL;
     ssize_t n_read;
     gboolean auth_complete;
 
@@ -483,13 +487,9 @@ from_child_cb (GIOChannel *source, GIOCondition condition, gpointer data)
     if (g_strcmp0 (username, session->priv->username) != 0)
     {
         g_free (session->priv->username);
-        session->priv->username = username;
-        if (session->priv->user)
-            g_object_unref (session->priv->user);
-        session->priv->user = NULL;
+        session->priv->username = g_steal_pointer (&username);
+        g_clear_object (&session->priv->user);
     }
-    else
-        g_free (username);
 
     /* Check if authentication completed */
     n_read = read_from_child (session, &auth_complete, sizeof (auth_complete));
@@ -552,13 +552,22 @@ session_get_is_started (Session *session)
     return session->priv->pid != 0;
 }
 
+static Greeter *
+create_greeter_cb (GreeterSocket *socket, Session *session)
+{
+    Greeter *greeter;
+    g_signal_emit (session, signals[CREATE_GREETER], 0, &greeter);
+    return greeter;
+}
+
 static gboolean
 session_real_start (Session *session)
 {
     int version;
     int to_child_pipe[2], from_child_pipe[2];
     int to_child_output, from_child_input;
-    gchar *arg0, *arg1;
+    g_autofree gchar *arg0 = NULL;
+    g_autofree gchar *arg1 = NULL;
 
     g_return_val_if_fail (session->priv->pid == 0, FALSE);
 
@@ -603,8 +612,6 @@ session_real_start (Session *session)
                 arg0, arg1, NULL);
         _exit (EXIT_FAILURE);
     }
-    g_free (arg0);
-    g_free (arg1);
 
     if (session->priv->pid < 0)
     {
@@ -626,7 +633,7 @@ session_real_start (Session *session)
     close (from_child_input);
 
     /* Indicate what version of the protocol we are using */
-    version = 2;
+    version = 3;
     write_data (session, &version, sizeof (version));
 
     /* Send configuration */
@@ -652,6 +659,13 @@ session_get_username (Session *session)
     return session->priv->username;
 }
 
+const gchar *
+session_get_login1_session_id (Session *session)
+{
+    g_return_val_if_fail (session != NULL, NULL);
+    return session->priv->login1_session_id;
+}
+
 const gchar *
 session_get_console_kit_cookie (Session *session)
 {
@@ -729,15 +743,24 @@ session_get_authentication_result_string (Session *session)
 void
 session_run (Session *session)
 {
+    g_return_if_fail (session != NULL);
     g_return_if_fail (session->priv->display_server != NULL);
     return SESSION_GET_CLASS (session)->run (session);
 }
 
+gboolean
+session_get_is_run (Session *session)
+{
+    g_return_val_if_fail (session != NULL, FALSE);
+    return session->priv->command_run;
+}
+
 static void
 session_real_run (Session *session)
 {
     gsize i, argc;
-    gchar *command, *x_authority_filename;
+    g_autofree gchar *command = NULL;
+    g_autofree gchar *x_authority_filename = NULL;
     GList *link;
 
     g_return_if_fail (session != NULL);
@@ -752,16 +775,15 @@ session_real_run (Session *session)
 
     command = g_strjoinv (" ", session->priv->argv);
     l_debug (session, "Running command %s", command);
-    g_free (command);
 
     /* Create authority location */
     if (session->priv->x_authority_use_system_location)
     {
-        gchar *run_dir, *dir;
+        g_autofree gchar *run_dir = NULL;
+        g_autofree gchar *dir = NULL;
 
         run_dir = config_get_string (config_get_instance (), "LightDM", "run-directory");
         dir = g_build_filename (run_dir, session->priv->username, NULL);
-        g_free (run_dir);
 
         if (g_mkdir_with_parents (dir, S_IRWXU) < 0)
             l_warning (session, "Failed to set create system authority dir %s: %s", dir, strerror (errno));
@@ -772,7 +794,6 @@ session_real_run (Session *session)
         }
 
         x_authority_filename = g_build_filename (dir, "xauthority", NULL);
-        g_free (dir);
     }
     else
         x_authority_filename = g_build_filename (user_get_home_directory (session_get_user (session)), ".Xauthority", NULL);
@@ -780,20 +801,45 @@ session_real_run (Session *session)
     /* Make sure shared user directory for this user exists */
     if (!session->priv->remote_host_name)
     {
-        gchar *data_dir = shared_data_manager_ensure_user_dir (shared_data_manager_get_instance (), session->priv->username);
+        g_autofree gchar *data_dir = shared_data_manager_ensure_user_dir (shared_data_manager_get_instance (), session->priv->username);
         if (data_dir)
-        {
             session_set_env (session, "XDG_GREETER_DATA_DIR", data_dir);
-            g_free (data_dir);
+    }
+
+    /* Open socket to allow in-session greeter */
+    if (session->priv->config && session_config_get_allow_greeter (session->priv->config))
+    {
+        g_autofree gchar *run_dir = NULL;
+        g_autofree gchar *dir = NULL;
+        g_autofree gchar *path = NULL;
+        g_autoptr(GError) error = NULL;
+
+        run_dir = config_get_string (config_get_instance (), "LightDM", "run-directory");
+        dir = g_build_filename (run_dir, session->priv->username, NULL);
+
+        if (g_mkdir_with_parents (dir, S_IRWXU) < 0)
+            l_warning (session, "Failed to create greeter socket dir %s: %s", dir, strerror (errno));
+        if (getuid () == 0)
+        {
+            if (chown (dir, user_get_uid (session_get_user (session)), user_get_gid (session_get_user (session))) < 0)
+                l_warning (session, "Failed to set ownership of greeter socket dir: %s", strerror (errno));
         }
+
+        path = g_build_filename (dir, "greeter-socket", NULL);
+        session->priv->greeter_socket = greeter_socket_new (path);
+        g_signal_connect (session->priv->greeter_socket, GREETER_SOCKET_SIGNAL_CREATE_GREETER, G_CALLBACK (create_greeter_cb), session);
+        session_set_env (session, "LIGHTDM_GREETER_PIPE", path);
+
+        if (!greeter_socket_start (session->priv->greeter_socket, &error))
+            l_warning (session, "Failed to start greeter socket: %s\n", error->message);
     }
 
     if (session->priv->log_filename)
         l_debug (session, "Logging to %s", session->priv->log_filename);
     write_string (session, session->priv->log_filename);
+    write_data (session, &session->priv->log_mode, sizeof (session->priv->log_mode));
     write_string (session, session->priv->tty);
     write_string (session, x_authority_filename);
-    g_free (x_authority_filename);
     write_string (session, session->priv->xdisplay);
     write_xauth (session, session->priv->x_authority);
     argc = g_list_length (session->priv->env);
@@ -805,7 +851,7 @@ session_real_run (Session *session)
     for (i = 0; i < argc; i++)
         write_string (session, session->priv->argv[i]);
 
-    session->priv->login1_session = read_string_from_child (session);
+    session->priv->login1_session_id = read_string_from_child (session);
     session->priv->console_kit_cookie = read_string_from_child (session);
 }
 
@@ -815,8 +861,8 @@ session_lock (Session *session)
     g_return_if_fail (session != NULL);
     if (getuid () == 0)
     {
-        if (session->priv->login1_session)
-            login1_service_lock_session (login1_service_get_instance (), session->priv->login1_session);
+        if (session->priv->login1_session_id)
+            login1_service_lock_session (login1_service_get_instance (), session->priv->login1_session_id);
         else if (session->priv->console_kit_cookie)
             ck_lock_session (session->priv->console_kit_cookie);
     }
@@ -828,8 +874,8 @@ session_unlock (Session *session)
     g_return_if_fail (session != NULL);
     if (getuid () == 0)
     {
-        if (session->priv->login1_session)
-            login1_service_unlock_session (login1_service_get_instance (), session->priv->login1_session);
+        if (session->priv->login1_session_id)
+            login1_service_unlock_session (login1_service_get_instance (), session->priv->login1_session_id);
         else if (session->priv->console_kit_cookie)
             ck_unlock_session (session->priv->console_kit_cookie);
     }
@@ -841,8 +887,8 @@ session_activate (Session *session)
     g_return_if_fail (session != NULL);
     if (getuid () == 0)
     {
-        if (session->priv->login1_session)
-            login1_service_activate_session (login1_service_get_instance (), session->priv->login1_session);
+        if (session->priv->login1_session_id)
+            login1_service_activate_session (login1_service_get_instance (), session->priv->login1_session_id);
         else if (session->priv->console_kit_cookie)
             ck_activate_session (session->priv->console_kit_cookie);
     }
@@ -853,13 +899,24 @@ session_stop (Session *session)
 {
     g_return_if_fail (session != NULL);
 
+    if (session->priv->stopping)
+        return;
+    session->priv->stopping = TRUE;
+
+    /* Kill remaining processes in our logind session to avoid them leaking
+     * to the user session (they share the same $DISPLAY) */
+    if (getuid () == 0 && session->priv->login1_session_id)
+        login1_service_terminate_session (login1_service_get_instance (), session->priv->login1_session_id);
+
     /* If can cleanly stop then do that */
     if (session_get_is_authenticated (session) && !session->priv->command_run)
     {
         gsize n = 0;
+        LogMode log_mode = LOG_MODE_INVALID;
 
         session->priv->command_run = TRUE;
         write_string (session, NULL); // log filename
+        write_data (session, &log_mode, sizeof (log_mode)); // log mode
         write_string (session, NULL); // tty
         write_string (session, NULL); // xauth filename
         write_string (session, NULL); // xdisplay
@@ -869,10 +926,6 @@ session_stop (Session *session)
         return;
     }
 
-    if (session->priv->stopping)
-        return;
-    session->priv->stopping = TRUE;
-
     return SESSION_GET_CLASS (session)->stop (session);
 }
 
@@ -903,6 +956,9 @@ session_init (Session *session)
 {
     session->priv = G_TYPE_INSTANCE_GET_PRIVATE (session, SESSION_TYPE, SessionPrivate);
     session->priv->log_filename = g_strdup (".xsession-errors");
+    session->priv->log_mode = LOG_MODE_BACKUP_AND_TRUNCATE;
+    session->priv->to_child_input = -1;
+    session->priv->from_child_output = -1;
 }
 
 static void
@@ -911,36 +967,33 @@ session_finalize (GObject *object)
     Session *self = SESSION (object);
     int i;
 
-    if (self->priv->config)
-        g_object_unref (self->priv->config);
-    if (self->priv->display_server)
-        g_object_unref (self->priv->display_server);
+    g_clear_object (&self->priv->config);
+    g_clear_object (&self->priv->display_server);
     if (self->priv->pid)
         kill (self->priv->pid, SIGKILL);
-    if (self->priv->from_child_channel)
-        g_io_channel_unref (self->priv->from_child_channel);
+    close (self->priv->to_child_input);
+    close (self->priv->from_child_output);
+    g_clear_pointer (&self->priv->from_child_channel, g_io_channel_unref);
     if (self->priv->from_child_watch)
         g_source_remove (self->priv->from_child_watch);
     if (self->priv->child_watch)
         g_source_remove (self->priv->child_watch);
-    g_free (self->priv->username);
-    if (self->priv->user)
-        g_object_unref (self->priv->user);
-    g_free (self->priv->pam_service);
+    g_clear_pointer (&self->priv->username, g_free);
+    g_clear_object (&self->priv->user);
+    g_clear_pointer (&self->priv->pam_service, g_free);
     for (i = 0; i < self->priv->messages_length; i++)
         g_free ((char *) self->priv->messages[i].msg);
-    g_free (self->priv->messages);
-    g_free (self->priv->authentication_result_string);
-    g_free (self->priv->log_filename);
-    g_free (self->priv->tty);
-    g_free (self->priv->xdisplay);
-    if (self->priv->x_authority)
-        g_object_unref (self->priv->x_authority);
-    g_free (self->priv->remote_host_name);
-    g_free (self->priv->login1_session);
-    g_free (self->priv->console_kit_cookie);
+    g_clear_pointer (&self->priv->messages, g_free);
+    g_clear_pointer (&self->priv->authentication_result_string, g_free);
+    g_clear_pointer (&self->priv->log_filename, g_free);
+    g_clear_pointer (&self->priv->tty, g_free);
+    g_clear_pointer (&self->priv->xdisplay, g_free);
+    g_clear_object (&self->priv->x_authority);
+    g_clear_pointer (&self->priv->remote_host_name, g_free);
+    g_clear_pointer (&self->priv->login1_session_id, g_free);
+    g_clear_pointer (&self->priv->console_kit_cookie, g_free);
     g_list_free_full (self->priv->env, g_free);
-    g_strfreev (self->priv->argv);
+    g_clear_pointer (&self->priv->argv, g_strfreev);
 
     G_OBJECT_CLASS (session_parent_class)->finalize (object);
 }
@@ -957,8 +1010,18 @@ session_class_init (SessionClass *klass)
 
     g_type_class_add_private (klass, sizeof (SessionPrivate));
 
+    signals[CREATE_GREETER] =
+        g_signal_new (SESSION_SIGNAL_CREATE_GREETER,
+                      G_TYPE_FROM_CLASS (klass),
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET (SessionClass, create_greeter),
+                      g_signal_accumulator_first_wins,
+                      NULL,
+                      NULL,
+                      GREETER_TYPE, 0);
+
     signals[GOT_MESSAGES] =
-        g_signal_new ("got-messages",
+        g_signal_new (SESSION_SIGNAL_GOT_MESSAGES,
                       G_TYPE_FROM_CLASS (klass),
                       G_SIGNAL_RUN_LAST,
                       G_STRUCT_OFFSET (SessionClass, got_messages),
@@ -967,7 +1030,7 @@ session_class_init (SessionClass *klass)
                       G_TYPE_NONE, 0);
 
     signals[AUTHENTICATION_COMPLETE] =
-        g_signal_new ("authentication-complete",
+        g_signal_new (SESSION_SIGNAL_AUTHENTICATION_COMPLETE,
                       G_TYPE_FROM_CLASS (klass),
                       G_SIGNAL_RUN_LAST,
                       G_STRUCT_OFFSET (SessionClass, authentication_complete),
@@ -976,7 +1039,7 @@ session_class_init (SessionClass *klass)
                       G_TYPE_NONE, 0);
 
     signals[STOPPED] =
-        g_signal_new ("stopped",
+        g_signal_new (SESSION_SIGNAL_STOPPED,
                       G_TYPE_FROM_CLASS (klass),
                       G_SIGNAL_RUN_LAST,
                       G_STRUCT_OFFSET (SessionClass, stopped),
