@@ -1,7 +1,7 @@
 /*
  * Copyright (C) 2010-2011 Robert Ancell.
  * Author: Robert Ancell <robert.ancell@canonical.com>
- * 
+ *
  * This program is free software: you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free Software
  * Foundation, either version 3 of the License, or (at your option) any later
@@ -18,11 +18,19 @@
 
 typedef struct Session Session;
 
+typedef enum
+{
+    SESSION_TYPE_LOCAL,
+    SESSION_TYPE_REMOTE
+} SessionType;
+
 #include "session-config.h"
 #include "display-server.h"
 #include "accounts.h"
 #include "x-authority.h"
 #include "logger.h"
+#include "log-file.h"
+#include "greeter.h"
 
 G_BEGIN_DECLS
 
@@ -31,6 +39,11 @@ G_BEGIN_DECLS
 #define SESSION_CLASS(klass)   (G_TYPE_CHECK_CLASS_CAST ((klass), SESSION_TYPE, SessionClass))
 #define SESSION_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), SESSION_TYPE, SessionClass))
 
+#define SESSION_SIGNAL_CREATE_GREETER          "create-greeter"
+#define SESSION_SIGNAL_GOT_MESSAGES            "got-messages"
+#define SESSION_SIGNAL_AUTHENTICATION_COMPLETE "authentication-complete"
+#define SESSION_SIGNAL_STOPPED                 "stopped"
+
 typedef struct SessionPrivate SessionPrivate;
 
 struct Session
@@ -47,16 +60,13 @@ typedef struct
     void (*run)(Session *session);
     void (*stop)(Session *session);
 
+    Greeter *(*create_greeter)(Session *session);
     void (*got_messages)(Session *session);
     void (*authentication_complete)(Session *session);
     void (*stopped)(Session *session);
 } SessionClass;
 
-typedef enum
-{
-    SESSION_TYPE_LOCAL,
-    SESSION_TYPE_REMOTE
-} SessionType;
+G_DEFINE_AUTOPTR_CLEANUP_FUNC (Session, g_object_unref);
 
 GType session_get_type (void);
 
@@ -80,7 +90,7 @@ void session_set_is_guest (Session *session, gboolean is_guest);
 
 gboolean session_get_is_guest (Session *session);
 
-void session_set_log_file (Session *session, const gchar *filename);
+void session_set_log_file (Session *session, const gchar *filename, LogMode log_mode);
 
 void session_set_display_server (Session *session, DisplayServer *display_server);
 
@@ -111,6 +121,8 @@ gboolean session_get_is_started (Session *session);
 
 const gchar *session_get_username (Session *session);
 
+const gchar *session_get_login1_session_id (Session *session);
+
 const gchar *session_get_console_kit_cookie (Session *session);
 
 void session_respond (Session *session, struct pam_response *response);
@@ -129,6 +141,8 @@ const gchar *session_get_authentication_result_string (Session *session);
 
 void session_run (Session *session);
 
+gboolean session_get_is_run (Session *session);
+
 void session_lock (Session *session);
 
 void session_unlock (Session *session);
