@@ -1,7 +1,7 @@
 /*
  * Copyright (C) 2013 Canonical Ltd.
  * Author: Robert Ancell <robert.ancell@canonical.com>
- * 
+ *
  * This program is free software: you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free Software
  * Foundation, either version 3 of the License, or (at your option) any later
@@ -21,7 +21,7 @@
 #include "unity-system-compositor.h"
 #include "configuration.h"
 #include "process.h"
-#include "greeter.h"
+#include "greeter-session.h"
 #include "vt.h"
 
 struct UnitySystemCompositorPrivate
@@ -39,9 +39,6 @@ struct UnitySystemCompositorPrivate
     gint vt;
     gboolean have_vt_ref;
 
-    /* TRUE if should show hardware cursor */
-    gboolean enable_hardware_cursor;
-
     /* Pipes to communicate with compositor */
     int to_compositor_pipe[2];
     int from_compositor_pipe[2];
@@ -61,9 +58,16 @@ struct UnitySystemCompositorPrivate
 
     /* TRUE when received ready signal */
     gboolean is_ready;
+
+    /* Counters for Mir IDs to use */
+    int next_session_id;
+    int next_greeter_id;  
 };
 
-G_DEFINE_TYPE (UnitySystemCompositor, unity_system_compositor, DISPLAY_SERVER_TYPE);
+static void unity_system_compositor_logger_iface_init (LoggerInterface *iface);
+
+G_DEFINE_TYPE_WITH_CODE (UnitySystemCompositor, unity_system_compositor, DISPLAY_SERVER_TYPE,
+                         G_IMPLEMENT_INTERFACE (LOGGER_TYPE, unity_system_compositor_logger_iface_init));               
 
 typedef enum
 {
@@ -85,6 +89,8 @@ void
 unity_system_compositor_set_command (UnitySystemCompositor *compositor, const gchar *command)
 {
     g_return_if_fail (compositor != NULL);
+    g_return_if_fail (command != NULL);
+
     g_free (compositor->priv->command);
     compositor->priv->command = g_strdup (command);
 }
@@ -120,13 +126,6 @@ unity_system_compositor_set_vt (UnitySystemCompositor *compositor, gint vt)
     }
 }
 
-void
-unity_system_compositor_set_enable_hardware_cursor (UnitySystemCompositor *compositor, gboolean enable_cursor)
-{
-    g_return_if_fail (compositor != NULL);
-    compositor->priv->enable_hardware_cursor = enable_cursor;
-}
-
 void
 unity_system_compositor_set_timeout (UnitySystemCompositor *compositor, gint timeout)
 {
@@ -137,7 +136,7 @@ unity_system_compositor_set_timeout (UnitySystemCompositor *compositor, gint tim
 static void
 write_message (UnitySystemCompositor *compositor, guint16 id, const guint8 *payload, guint16 payload_length)
 {
-    guint8 *data;
+    g_autofree guint8 *data = NULL;
     gsize data_length = 4 + payload_length;
 
     data = g_malloc (data_length);
@@ -145,13 +144,12 @@ write_message (UnitySystemCompositor *compositor, guint16 id, const guint8 *payl
     data[1] = id & 0xFF;
     data[2] = payload_length >> 8;
     data[3] = payload_length & 0xFF;
-    memcpy (data + 4, payload, payload_length);
+    if (payload)
+        memcpy (data + 4, payload, payload_length);
 
     errno = 0;
     if (write (compositor->priv->to_compositor_pipe[1], data, data_length) != data_length)
         l_warning (compositor, "Failed to write to compositor: %s", strerror (errno));
-
-    g_free (data);
 }
 
 void
@@ -179,28 +177,41 @@ static void
 unity_system_compositor_connect_session (DisplayServer *display_server, Session *session)
 {
     UnitySystemCompositor *compositor = UNITY_SYSTEM_COMPOSITOR (display_server);
-    const gchar *name;
+
+    session_set_env (session, "XDG_SESSION_TYPE", "mir");
 
     if (compositor->priv->socket)
-        session_set_env (session, "MIR_SOCKET", compositor->priv->socket);
-    if (IS_GREETER (session))
-        name = "greeter-0";
-    else
-        name = "session-0";
-    session_set_env (session, "MIR_SERVER_NAME", name);
+        session_set_env (session, "MIR_SERVER_HOST_SOCKET", compositor->priv->socket);
+
+    if (!session_get_env (session, "MIR_SERVER_NAME"))
+    {
+        g_autofree gchar *name = NULL;
+
+        if (IS_GREETER_SESSION (session))
+        {
+            name = g_strdup_printf ("greeter-%d", compositor->priv->next_greeter_id);
+            compositor->priv->next_greeter_id++;
+        }
+        else
+        {
+            name = g_strdup_printf ("session-%d", compositor->priv->next_session_id);
+            compositor->priv->next_session_id++;
+        }
+        session_set_env (session, "MIR_SERVER_NAME", name);
+    }
 
     if (compositor->priv->vt >= 0)
     {
-        gchar *value = g_strdup_printf ("%d", compositor->priv->vt);
+        g_autofree gchar *value = g_strdup_printf ("%d", compositor->priv->vt);
         session_set_env (session, "XDG_VTNR", value);
-        g_free (value);
     }
 }
 
 static void
 unity_system_compositor_disconnect_session (DisplayServer *display_server, Session *session)
 {
-    session_unset_env (session, "MIR_SOCKET");
+    session_unset_env (session, "XDG_SESSION_TYPE");
+    session_unset_env (session, "MIR_SERVER_HOST_SOCKET");
     session_unset_env (session, "MIR_SERVER_NAME");
     session_unset_env (session, "XDG_VTNR");
 }
@@ -208,8 +219,9 @@ unity_system_compositor_disconnect_session (DisplayServer *display_server, Sessi
 static gchar *
 get_absolute_command (const gchar *command)
 {
-    gchar **tokens;
-    gchar *absolute_binary, *absolute_command = NULL;
+    g_auto(GStrv) tokens = NULL;
+    g_autofree gchar *absolute_binary = NULL;
+    gchar *absolute_command = NULL;
 
     tokens = g_strsplit (command, " ", 2);
 
@@ -221,9 +233,6 @@ get_absolute_command (const gchar *command)
         else
             absolute_command = g_strdup (absolute_binary);
     }
-    g_free (absolute_binary);
-
-    g_strfreev (tokens);
 
     return absolute_command;
 }
@@ -235,7 +244,7 @@ read_cb (GIOChannel *source, GIOCondition condition, gpointer data)
     gsize n_to_read = 0;
     guint16 id, payload_length;
     /*guint8 *payload;*/
-  
+
     if (condition == G_IO_HUP)
     {
         l_debug (compositor, "Compositor closed communication channel");
@@ -257,7 +266,7 @@ read_cb (GIOChannel *source, GIOCondition condition, gpointer data)
     {
         gsize n_total, n_read = 0;
         GIOStatus status;
-        GError *error = NULL;
+        g_autoptr(GError) error = NULL;
 
         n_total = compositor->priv->read_buffer_n_used + n_to_read;
         if (compositor->priv->read_buffer_length < n_total)
@@ -272,7 +281,6 @@ read_cb (GIOChannel *source, GIOCondition condition, gpointer data)
             l_warning (compositor, "Failed to read from compositor: %s", error->message);
         if (status != G_IO_STATUS_NORMAL)
             return TRUE;
-        g_clear_error (&error);
         compositor->priv->read_buffer_n_used += n_read;
     }
 
@@ -368,9 +376,12 @@ static gboolean
 unity_system_compositor_start (DisplayServer *server)
 {
     UnitySystemCompositor *compositor = UNITY_SYSTEM_COMPOSITOR (server);
-    gboolean result;
-    GString *command;
-    gchar *dir, *log_file, *absolute_command, *value;
+    gboolean result, backup_logs;
+    g_autoptr(GString) command = NULL;
+    g_autofree gchar *dir = NULL;
+    g_autofree gchar *log_file = NULL;
+    g_autofree gchar *value = NULL;
+    g_autofree gchar *absolute_command = NULL;
 
     g_return_val_if_fail (compositor->priv->process == NULL, FALSE);
 
@@ -397,17 +408,15 @@ unity_system_compositor_start (DisplayServer *server)
     dir = config_get_string (config_get_instance (), "LightDM", "log-directory");
     log_file = g_build_filename (dir, "unity-system-compositor.log", NULL);
     l_debug (compositor, "Logging to %s", log_file);
-    g_free (dir);
 
     /* Setup environment */
     compositor->priv->process = process_new (run_cb, compositor);
-    process_set_log_file (compositor->priv->process, log_file, TRUE);
-    g_free (log_file);
+    backup_logs = config_get_boolean (config_get_instance (), "LightDM", "backup-logs");
+    process_set_log_file (compositor->priv->process, log_file, TRUE, backup_logs ? LOG_MODE_BACKUP_AND_TRUNCATE : LOG_MODE_APPEND);
     process_set_clear_environment (compositor->priv->process, TRUE);
     process_set_env (compositor->priv->process, "XDG_SEAT", "seat0");
     value = g_strdup_printf ("%d", compositor->priv->vt);
     process_set_env (compositor->priv->process, "XDG_VTNR", value);
-    g_free (value);
     /* Variable required for regression tests */
     if (g_getenv ("LIGHTDM_TEST_ROOT"))
     {
@@ -424,25 +433,21 @@ unity_system_compositor_start (DisplayServer *server)
         return FALSE;
     }
     command = g_string_new (absolute_command);
-    g_free (absolute_command);
     g_string_append_printf (command, " --file '%s'", compositor->priv->socket);
     g_string_append_printf (command, " --from-dm-fd %d --to-dm-fd %d", compositor->priv->to_compositor_pipe[0], compositor->priv->from_compositor_pipe[1]);
     if (compositor->priv->vt > 0)
         g_string_append_printf (command, " --vt %d", compositor->priv->vt);
-    if (compositor->priv->enable_hardware_cursor)
-        g_string_append (command, " --enable-hardware-cursor=true");
     process_set_command (compositor->priv->process, command->str);
-    g_string_free (command, TRUE);
 
     /* Start the compositor */
-    g_signal_connect (compositor->priv->process, "stopped", G_CALLBACK (stopped_cb), compositor);
+    g_signal_connect (compositor->priv->process, PROCESS_SIGNAL_STOPPED, G_CALLBACK (stopped_cb), compositor);
     result = process_start (compositor->priv->process, FALSE);
 
     /* Close compostor ends of the pipes */
     close (compositor->priv->to_compositor_pipe[0]);
-    compositor->priv->to_compositor_pipe[0] = 0;
+    compositor->priv->to_compositor_pipe[0] = -1;
     close (compositor->priv->from_compositor_pipe[1]);
-    compositor->priv->from_compositor_pipe[1] = 0;
+    compositor->priv->from_compositor_pipe[1] = -1;
 
     if (!result)
         return FALSE;
@@ -456,7 +461,7 @@ unity_system_compositor_start (DisplayServer *server)
 
     return TRUE;
 }
- 
+
 static void
 unity_system_compositor_stop (DisplayServer *server)
 {
@@ -469,34 +474,34 @@ unity_system_compositor_init (UnitySystemCompositor *compositor)
     compositor->priv = G_TYPE_INSTANCE_GET_PRIVATE (compositor, UNITY_SYSTEM_COMPOSITOR_TYPE, UnitySystemCompositorPrivate);
     compositor->priv->vt = -1;
     compositor->priv->command = g_strdup ("unity-system-compositor");
-    compositor->priv->socket = g_strdup ("/tmp/mir_socket");
+    compositor->priv->socket = g_strdup ("/run/mir_socket");
     compositor->priv->timeout = -1;
+    compositor->priv->to_compositor_pipe[0] = -1;
+    compositor->priv->to_compositor_pipe[1] = -1;
+    compositor->priv->from_compositor_pipe[0] = -1;
+    compositor->priv->from_compositor_pipe[1] = -1;
 }
 
 static void
 unity_system_compositor_finalize (GObject *object)
 {
-    UnitySystemCompositor *self;
-
-    self = UNITY_SYSTEM_COMPOSITOR (object);  
+    UnitySystemCompositor *self = UNITY_SYSTEM_COMPOSITOR (object);
 
     if (self->priv->process)
-    {
         g_signal_handlers_disconnect_matched (self->priv->process, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, self);
-        g_object_unref (self->priv->process);
-    }
-    g_free (self->priv->command);
-    g_free (self->priv->socket);
+    g_clear_object (&self->priv->process);
+    g_clear_pointer (&self->priv->command, g_free);
+    g_clear_pointer (&self->priv->socket, g_free);
     if (self->priv->have_vt_ref)
         vt_unref (self->priv->vt);
     close (self->priv->to_compositor_pipe[0]);
     close (self->priv->to_compositor_pipe[1]);
     close (self->priv->from_compositor_pipe[0]);
     close (self->priv->from_compositor_pipe[1]);
-    g_io_channel_unref (self->priv->from_compositor_channel);
-    if (self->priv->from_compositor_watch)      
+    g_clear_pointer (&self->priv->from_compositor_channel, g_io_channel_unref);
+    if (self->priv->from_compositor_watch)
         g_source_remove (self->priv->from_compositor_watch);
-    g_free (self->priv->read_buffer);
+    g_clear_pointer (&self->priv->read_buffer, g_free);
     if (self->priv->timeout_source)
         g_source_remove (self->priv->timeout_source);
 
@@ -518,3 +523,15 @@ unity_system_compositor_class_init (UnitySystemCompositorClass *klass)
 
     g_type_class_add_private (klass, sizeof (UnitySystemCompositorPrivate));
 }
+
+static gint
+unity_system_compositor_real_logprefix (Logger *self, gchar *buf, gulong buflen)
+{
+    return g_snprintf (buf, buflen, "Unity System Compositor: ");
+}
+
+static void
+unity_system_compositor_logger_iface_init (LoggerInterface *iface)
+{
+    iface->logprefix = &unity_system_compositor_real_logprefix;
+}
