@@ -1,7 +1,7 @@
 /*
  * Copyright (C) 2010-2011 Robert Ancell.
  * Author: Robert Ancell <robert.ancell@canonical.com>
- * 
+ *
  * This program is free software: you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free Software
  * Foundation, either version 3 of the License, or (at your option) any later
@@ -10,8 +10,10 @@
  */
 
 #include <string.h>
+#include <gio/gio.h>
 
 #include "xdmcp-protocol.h"
+#include "x-authority.h"
 
 typedef struct
 {
@@ -120,7 +122,7 @@ write_card16 (PacketWriter *writer, guint16 value)
 }
 
 static void
-write_card32 (PacketWriter *writer, guint16 value)
+write_card32 (PacketWriter *writer, guint32 value)
 {
     write_card8 (writer, (value >> 24) & 0xFF);
     write_card8 (writer, (value >> 16) & 0xFF);
@@ -187,7 +189,7 @@ xdmcp_packet_decode (const guint8 *data, gsize data_length)
     length = read_card16 (&reader);
 
     if (reader.overflow)
-    {      
+    {
         g_warning ("Ignoring short packet"); // FIXME: Use GError
         return NULL;
     }
@@ -211,8 +213,8 @@ xdmcp_packet_decode (const guint8 *data, gsize data_length)
         packet->Query.authentication_names = read_string_array (&reader);
         break;
     case XDMCP_ForwardQuery:
-        packet->ForwardQuery.client_address = read_string (&reader);
-        packet->ForwardQuery.client_port = read_string (&reader);
+        read_data (&reader, &packet->ForwardQuery.client_address);
+        read_data (&reader, &packet->ForwardQuery.client_port);
         packet->ForwardQuery.authentication_names = read_string_array (&reader);
         break;
     case XDMCP_Willing:
@@ -324,8 +326,8 @@ xdmcp_packet_encode (XDMCPPacket *packet, guint8 *data, gsize max_length)
         write_string_array (&writer, packet->Query.authentication_names);
         break;
     case XDMCP_ForwardQuery:
-        write_string (&writer, packet->ForwardQuery.client_address);
-        write_string (&writer, packet->ForwardQuery.client_port);
+        write_data (&writer, &packet->ForwardQuery.client_address);
+        write_data (&writer, &packet->ForwardQuery.client_port);
         write_string_array (&writer, packet->ForwardQuery.authentication_names);
         break;
     case XDMCP_Willing:
@@ -351,7 +353,7 @@ xdmcp_packet_encode (XDMCPPacket *packet, guint8 *data, gsize max_length)
         write_string (&writer, packet->Request.manufacturer_display_id);
         break;
     case XDMCP_Accept:
-        write_card32 (&writer, packet->Accept.session_id);      
+        write_card32 (&writer, packet->Accept.session_id);
         write_string (&writer, packet->Accept.authentication_name);
         write_data (&writer, &packet->Accept.authentication_data);
         write_string (&writer, packet->Accept.authorization_name);
@@ -399,118 +401,124 @@ xdmcp_packet_encode (XDMCPPacket *packet, guint8 *data, gsize max_length)
         g_warning ("Overflow writing response");
         return -1;
     }
-  
+
     return length + 6;
 }
 
 static gchar *
 data_tostring (XDMCPData *data)
 {
-    GString *s;
+    g_autoptr(GString) s = NULL;
     guint16 i;
-    gchar *string;
 
     s = g_string_new ("");
     for (i = 0; i < data->length; i++)
         g_string_append_printf (s, "%02X", data->data[i]);
-    string = s->str;
-    g_string_free (s, FALSE);
 
-    return string;
+    return g_steal_pointer (&s->str);
 }
 
 static gchar *
 string_list_tostring (gchar **strings)
 {
-    GString *s;
-    gchar *string;
+    g_autoptr(GString) s = NULL;
     gchar **i;
-  
-    s = g_string_new ("");  
+
+    s = g_string_new ("");
     for (i = strings; *i; i++)
     {
         if (i != strings)
            g_string_append (s, " ");
         g_string_append_printf (s, "'%s'", *i);
     }
-    string = s->str;
-    g_string_free (s, FALSE);
 
-    return string;
+    return g_steal_pointer (&s->str);
 }
 
 gchar *
 xdmcp_packet_tostring (XDMCPPacket *packet)
 {
-    gchar *string, *t, *t2;
-    gint i;
-    GString *t3;
-
     switch (packet->opcode)
     {
     case XDMCP_BroadcastQuery:
-        t = string_list_tostring (packet->Query.authentication_names);
-        string = g_strdup_printf ("BroadcastQuery(authentication_names=[%s])", t);
-        g_free (t);
-        return string;
+    {
+        g_autofree gchar *names = string_list_tostring (packet->Query.authentication_names);
+        return g_strdup_printf ("BroadcastQuery(authentication_names=[%s])", names);
+    }
     case XDMCP_Query:
-        t = string_list_tostring (packet->Query.authentication_names);
-        string = g_strdup_printf ("Query(authentication_names=[%s])", t);
-        g_free (t);
-        return string;
+    {
+        g_autofree gchar *names = string_list_tostring (packet->Query.authentication_names);
+        return g_strdup_printf ("Query(authentication_names=[%s])", names);
+    }
     case XDMCP_IndirectQuery:
-        t = string_list_tostring (packet->Query.authentication_names);
-        string = g_strdup_printf ("IndirectQuery(authentication_names=[%s])", t);
-        g_free (t);
-        return string;
+    {
+        g_autofree gchar *names = string_list_tostring (packet->Query.authentication_names);
+        return g_strdup_printf ("IndirectQuery(authentication_names=[%s])", names);
+    }
     case XDMCP_ForwardQuery:
-        t = string_list_tostring (packet->ForwardQuery.authentication_names);
-        string = g_strdup_printf ("ForwardQuery(client_address='%s' client_port='%s' authentication_names=[%s])",
-                                  packet->ForwardQuery.client_address, packet->ForwardQuery.client_port, t);
-        g_free (t);
-        return string;
+    {
+        g_autofree gchar *address_text = data_tostring (&packet->ForwardQuery.client_address);
+        g_autofree gchar *port_text = data_tostring (&packet->ForwardQuery.client_port);
+        g_autofree gchar *names_text = string_list_tostring (packet->ForwardQuery.authentication_names);
+        return g_strdup_printf ("ForwardQuery(client_address=%s client_port=%s authentication_names=[%s])",
+                                address_text, port_text, names_text);
+    }
     case XDMCP_Willing:
         return g_strdup_printf ("Willing(authentication_name='%s' hostname='%s' status='%s')",
                                 packet->Willing.authentication_name, packet->Willing.hostname, packet->Willing.status);
     case XDMCP_Unwilling:
         return g_strdup_printf ("Unwilling(hostname='%s' status='%s')",
-                                packet->Unwilling.hostname, packet->Unwilling.status);      
+                                packet->Unwilling.hostname, packet->Unwilling.status);
     case XDMCP_Request:
-        t = string_list_tostring (packet->Request.authorization_names);
-        t2 = data_tostring (&packet->Request.authentication_data);
-        t3 = g_string_new ("");
+    {
+        g_autofree gchar *names_text = string_list_tostring (packet->Request.authorization_names);
+        g_autofree gchar *data_text = data_tostring (&packet->Request.authentication_data);
+        g_autoptr(GString) connections_text = g_string_new ("");
+        gint i;
+
         for (i = 0; i < packet->Request.n_connections; i++)
         {
-            gchar *t4;
+            XDMCPConnection *connection = &packet->Request.connections[i];
+            GSocketFamily family = G_SOCKET_FAMILY_INVALID;
 
             if (i != 0)
-               g_string_append (t3, " ");
-            t4 = data_tostring (&packet->Request.connections[i].address);
-            g_string_append_printf (t3, "(%d, %s)", packet->Request.connections[i].type, t4);
-            g_free (t4);
+               g_string_append (connections_text, " ");
+
+            if (connection->type == XAUTH_FAMILY_INTERNET && connection->address.length == 4)
+                family = G_SOCKET_FAMILY_IPV4;
+            else if (connection->type == XAUTH_FAMILY_INTERNET6 && connection->address.length == 16)
+                family = G_SOCKET_FAMILY_IPV6;
+
+            if (family != G_SOCKET_FAMILY_INVALID)
+            {
+                g_autoptr(GInetAddress) address = g_inet_address_new_from_bytes (connection->address.data, family);
+                g_autofree gchar *text = g_inet_address_to_string (address);
+                g_string_append (connections_text, text);
+            }
+            else
+            {
+                g_autofree gchar *text = data_tostring (&connection->address);
+                g_string_append_printf (connections_text, "(%d, %s)", connection->type, text);
+            }
         }
-        string = g_strdup_printf ("Request(display_number=%d connections=[%s] authentication_name='%s' authentication_data=%s authorization_names=[%s] manufacturer_display_id='%s')",
-                                  packet->Request.display_number, t3->str, packet->Request.authentication_name, t2,
-                                  t, packet->Request.manufacturer_display_id);
-        g_free (t);
-        g_free (t2);
-        g_string_free (t3, TRUE);
-        return string;
+        return g_strdup_printf ("Request(display_number=%d connections=[%s] authentication_name='%s' authentication_data=%s authorization_names=[%s] manufacturer_display_id='%s')",
+                                packet->Request.display_number, connections_text->str, packet->Request.authentication_name, data_text,
+                                names_text, packet->Request.manufacturer_display_id);
+    }
     case XDMCP_Accept:
-        t = data_tostring (&packet->Accept.authentication_data);
-        t2 = data_tostring (&packet->Accept.authorization_data);
-        string =  g_strdup_printf ("Accept(session_id=%d authentication_name='%s' authentication_data=%s authorization_name='%s' authorization_data=%s)",
-                                   packet->Accept.session_id, packet->Accept.authentication_name, t,
-                                   packet->Accept.authorization_name, t2);
-        g_free (t);
-        g_free (t2);
-        return string;
+    {
+        g_autofree gchar *authentication_text = data_tostring (&packet->Accept.authentication_data);
+        g_autofree gchar *authorization_text = data_tostring (&packet->Accept.authorization_data);
+        return g_strdup_printf ("Accept(session_id=%d authentication_name='%s' authentication_data=%s authorization_name='%s' authorization_data=%s)",
+                                packet->Accept.session_id, packet->Accept.authentication_name, authentication_text,
+                                packet->Accept.authorization_name, authorization_text);
+    }
     case XDMCP_Decline:
-        t = data_tostring (&packet->Decline.authentication_data);
-        string = g_strdup_printf ("Decline(status='%s' authentication_name='%s' authentication_data=%s)",
-                                  packet->Decline.status, packet->Decline.authentication_name, t);
-        g_free (t);
-        return string;
+    {
+        g_autofree gchar *t = data_tostring (&packet->Decline.authentication_data);
+        return g_strdup_printf ("Decline(status='%s' authentication_name='%s' authentication_data=%s)",
+                                packet->Decline.status, packet->Decline.authentication_name, t);
+    }
     case XDMCP_Manage:
         return g_strdup_printf ("Manage(session_id=%d display_number=%d display_class='%s')",
                                 packet->Manage.session_id, packet->Manage.display_number, packet->Manage.display_class);
@@ -545,8 +553,8 @@ xdmcp_packet_free (XDMCPPacket *packet)
         g_strfreev (packet->Query.authentication_names);
         break;
     case XDMCP_ForwardQuery:
-        g_free (packet->ForwardQuery.client_address);
-        g_free (packet->ForwardQuery.client_port);
+        g_free (packet->ForwardQuery.client_address.data);
+        g_free (packet->ForwardQuery.client_port.data);      
         g_strfreev (packet->ForwardQuery.authentication_names);
         break;
     case XDMCP_Willing:
