@@ -3,16 +3,18 @@
 #include <string.h>
 #include <sys/types.h>
 #include <unistd.h>
-#include <signal.h>
+#include <sys/stat.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <gio/gio.h>
+#include <glib-unix.h>
 
 #include "status.h"
 #include "x-server.h"
 #include "x-authority.h"
 
 static GMainLoop *loop;
+static int exit_status = EXIT_SUCCESS;
 
 static GKeyFile *config;
 
@@ -22,6 +24,9 @@ static gchar *lock_path = NULL;
 /* Path to authority database to use */
 static gchar *auth_path = NULL;
 
+/* ID to use for test reporting */
+static gchar *id;
+
 /* Display number being served */
 static int display_number = 0;
 
@@ -33,49 +38,50 @@ cleanup (void)
 {
     if (lock_path)
         unlink (lock_path);
-    if (xserver)
-        g_object_unref (xserver);
+    g_clear_object (&xserver);
 }
 
 static void
 quit (int status)
 {
-    cleanup ();
-    exit (status);
+    exit_status = status;
+    g_main_loop_quit (loop);
 }
 
-static void
-signal_cb (int signum)
+static gboolean
+sighup_cb (gpointer user_data)
 {
-    if (signum == SIGHUP)
-    {
-        status_notify ("XSERVER-%d DISCONNECT-CLIENTS", display_number);
-    }
-    else
-    {
-        status_notify ("XSERVER-%d TERMINATE SIGNAL=%d", display_number, signum);
-        quit (EXIT_SUCCESS);
-    }
+    status_notify ("%s DISCONNECT-CLIENTS", id);
+    return TRUE;
+}
+
+static gboolean
+sigint_cb (gpointer user_data)
+{
+    status_notify ("%s TERMINATE SIGNAL=%d", id, SIGINT);
+    quit (EXIT_SUCCESS);
+    return TRUE;
+}
+
+static gboolean
+sigterm_cb (gpointer user_data)
+{
+    status_notify ("%s TERMINATE SIGNAL=%d", id, SIGTERM);
+    quit (EXIT_SUCCESS);
+    return TRUE;
 }
 
 static void
 client_connected_cb (XServer *server, XClient *client)
 {
-    gchar *auth_error = NULL;
-
-    status_notify ("XSERVER-%d ACCEPT-CONNECT", display_number);
-
-    if (auth_error)
-        x_client_send_failed (client, auth_error);
-    else
-        x_client_send_success (client);
-    g_free (auth_error);
+    status_notify ("%s ACCEPT-CONNECT", id);
+    x_client_send_success (client);
 }
 
 static void
 client_disconnected_cb (XServer *server, XClient *client)
-{  
-    g_signal_handlers_disconnect_matched (client, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, NULL);  
+{
+    g_signal_handlers_disconnect_matched (client, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, NULL);
 }
 
 static gboolean
@@ -84,73 +90,63 @@ vnc_data_cb (GIOChannel *channel, GIOCondition condition, gpointer data)
     gchar buffer[1024];
     gsize n_read;
     GIOStatus status;
-    GError *error = NULL;
+    g_autoptr(GError) error = NULL;
 
     status = g_io_channel_read_chars (channel, buffer, 1023, &n_read, &error);
     if (error)
         g_warning ("Error reading from VNC client: %s", error->message);
-    g_clear_error (&error);
 
     if (status == G_IO_STATUS_NORMAL)
     {
         buffer[n_read] = '\0';
         if (g_str_has_suffix (buffer, "\n"))
             buffer[n_read-1] = '\0';
-        status_notify ("XSERVER-%d VNC-CLIENT-CONNECT VERSION=\"%s\"", display_number, buffer);
+        status_notify ("%s VNC-CLIENT-CONNECT VERSION=\"%s\"", id, buffer);
     }
-  
+
     return TRUE;
 }
 
 static void
-request_cb (const gchar *request)
+request_cb (const gchar *name, GHashTable *params)
 {
-    gchar *r;
-
-    if (!request)
+    if (!name)
     {
         g_main_loop_quit (loop);
         return;
     }
 
-    r = g_strdup_printf ("XSERVER-%d INDICATE-READY", display_number);
-    if (strcmp (request, r) == 0)
+    if (strcmp (name, "INDICATE-READY") == 0)
     {
         void *handler;
 
         handler = signal (SIGUSR1, SIG_IGN);
         if (handler == SIG_IGN)
         {
-            status_notify ("XSERVER-%d INDICATE-READY", display_number);
+            status_notify ("%s INDICATE-READY", id);
             kill (getppid (), SIGUSR1);
         }
         signal (SIGUSR1, handler);
     }
-    g_free (r);
-    r = g_strdup_printf ("XSERVER-%d START-VNC", display_number);
-    if (strcmp (request, r) == 0)
+
+    else if (strcmp (name, "START-VNC") == 0)
     {
         /* Send server protocol version to client */
         g_print ("RFB 003.007\n");
     }
-    g_free (r);
 }
 
 int
 main (int argc, char **argv)
 {
-    char *pid_string;
+    int i;
+    g_autofree gchar *pid_string = NULL;
     gboolean use_inetd = FALSE;
     gboolean has_option = FALSE;
-    gchar *geometry = g_strdup ("640x480");
+    const gchar *geometry = "640x480";
     gint depth = 8;
-    gchar *lock_filename;
+    g_autofree gchar *lock_filename = NULL;
     int lock_file;
-    int i;
-
-    signal (SIGINT, signal_cb);
-    signal (SIGTERM, signal_cb);
-    signal (SIGHUP, signal_cb);
 
 #if !defined(GLIB_VERSION_2_36)
     g_type_init ();
@@ -158,7 +154,9 @@ main (int argc, char **argv)
 
     loop = g_main_loop_new (NULL, FALSE);
 
-    status_connect (request_cb);
+    g_unix_signal_add (SIGINT, sigint_cb, NULL);
+    g_unix_signal_add (SIGTERM, sigterm_cb, NULL);
+    g_unix_signal_add (SIGHUP, sighup_cb, NULL);
 
     for (i = 1; i < argc; i++)
     {
@@ -184,9 +182,8 @@ main (int argc, char **argv)
         }
         else if (strcmp (arg, "-geometry") == 0)
         {
-            g_free (geometry);
-            geometry = g_strdup (argv[i+1]);
-            i++;         
+            geometry = argv[i+1];
+            i++;
         }
         else if (strcmp (arg, "-depth") == 0)
         {
@@ -214,16 +211,20 @@ main (int argc, char **argv)
             return EXIT_FAILURE;
         }
     }
-  
-    xserver = x_server_new (display_number);
-    g_signal_connect (xserver, "client-connected", G_CALLBACK (client_connected_cb), NULL);
-    g_signal_connect (xserver, "client-disconnected", G_CALLBACK (client_disconnected_cb), NULL);
 
-    status_notify ("XSERVER-%d START GEOMETRY=%s DEPTH=%d OPTION=%s", display_number, geometry, depth, has_option ? "TRUE" : "FALSE");
+    id = g_strdup_printf ("XVNC-%d", display_number);
+
+    status_connect (request_cb, id);
+
+    xserver = x_server_new (display_number);
+    g_signal_connect (xserver, X_SERVER_SIGNAL_CLIENT_CONNECTED, G_CALLBACK (client_connected_cb), NULL);
+    g_signal_connect (xserver, X_SERVER_SIGNAL_CLIENT_DISCONNECTED, G_CALLBACK (client_disconnected_cb), NULL);
+
+    status_notify ("%s START GEOMETRY=%s DEPTH=%d OPTION=%s", id, geometry, depth, has_option ? "TRUE" : "FALSE");
 
     config = g_key_file_new ();
     g_key_file_load_from_file (config, g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "script", NULL), G_KEY_FILE_NONE, NULL);
-  
+
     if (use_inetd)
     {
         if (!g_io_add_watch (g_io_channel_unix_new (STDIN_FILENO), G_IO_IN, vnc_data_cb, NULL))
@@ -237,33 +238,62 @@ main (int argc, char **argv)
 
     lock_filename = g_strdup_printf (".X%d-lock", display_number);
     lock_path = g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "tmp", lock_filename, NULL);
-    g_free (lock_filename);
     lock_file = open (lock_path, O_CREAT | O_EXCL | O_WRONLY, 0444);
     if (lock_file < 0)
+    {
+        g_autofree gchar *lock_contents = NULL;
+
+        if (g_file_get_contents (lock_path, &lock_contents, NULL, NULL))
+        {
+            g_autofree gchar *proc_filename = NULL;
+            pid_t pid;
+
+            pid = atol (lock_contents);
+
+            proc_filename = g_strdup_printf ("/proc/%d", pid);
+            if (!g_file_test (proc_filename, G_FILE_TEST_EXISTS))
+            {
+                g_autofree gchar *socket_dir = NULL;
+                g_autofree gchar *socket_filename = NULL;
+                g_autofree gchar *socket_path = NULL;
+
+                socket_dir = g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "tmp", ".X11-unix", NULL);
+                g_mkdir_with_parents (socket_dir, 0755);
+
+                socket_filename = g_strdup_printf ("X%d", display_number);
+                socket_path = g_build_filename (socket_dir, socket_filename, NULL);
+
+                g_printerr ("Breaking lock on non-existant process %d\n", pid);
+                unlink (lock_path);
+                unlink (socket_path);
+            }
+
+            lock_file = open (lock_path, O_CREAT | O_EXCL | O_WRONLY, 0444);
+        }
+    }
+    if (lock_file < 0)
     {
         fprintf (stderr,
                  "Fatal server error:\n"
                  "Server is already active for display %d\n"
                  "	If this server is no longer running, remove %s\n"
                  "	and start again.\n", display_number, lock_path);
-        g_free (lock_path);
-        lock_path = NULL;
-        quit (EXIT_FAILURE);
+        g_clear_pointer (&lock_path, g_free);
+        return EXIT_FAILURE;
     }
     pid_string = g_strdup_printf ("%10ld", (long) getpid ());
     if (write (lock_file, pid_string, strlen (pid_string)) < 0)
     {
         g_warning ("Error writing PID file: %s", strerror (errno));
-        quit (EXIT_FAILURE);
+        return EXIT_FAILURE;
     }
-    g_free (pid_string);
 
     if (!x_server_start (xserver))
-        quit (EXIT_FAILURE);
+        return EXIT_FAILURE;
 
     g_main_loop_run (loop);
 
     cleanup ();
 
-    return EXIT_SUCCESS;
+    return exit_status;
 }
