@@ -1,6 +1,8 @@
 #define _GNU_SOURCE
 #define __USE_GNU
 
+#include <config.h>
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -8,6 +10,9 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
+#include <sys/socket.h>
+#include <sys/un.h>
+#include <netinet/in.h>
 #include <pwd.h>
 #include <unistd.h>
 #include <dirent.h>
@@ -24,11 +29,15 @@
 #include <xcb/xcb.h>
 #include <gio/gunixsocketaddress.h>
 
+#if HAVE_LIBAUDIT
+#include <libaudit.h>
+#endif
+
 #include "status.h"
 
 #define LOGIN_PROMPT "login:"
 
-static int console_fd = -1;
+static int tty_fd = -1;
 
 static GList *user_entries = NULL;
 static GList *getpwent_link = NULL;
@@ -64,6 +73,13 @@ struct pam_handle
     struct pam_conv conversation;
 };
 
+int
+gethostname (char *name, size_t len)
+{
+   snprintf (name, len, "lightdm-test");
+   return 0;
+}
+
 uid_t
 getuid (void)
 {
@@ -91,7 +107,7 @@ int
 getgroups (int size, gid_t list[])
 {
     const gchar *group_list;
-    gchar **groups;
+    g_auto(GStrv) groups = NULL;
     gint groups_length;
 
     /* Get groups we are a member of */
@@ -113,7 +129,6 @@ getgroups (int size, gid_t list[])
         for (i = 0; groups[i]; i++)
             list[i] = atoi (groups[i]);
     }
-    g_free (groups);
 
     return groups_length;
 }
@@ -122,7 +137,7 @@ int
 setgroups (size_t size, const gid_t *list)
 {
     size_t i;
-    GString *group_list;
+    g_autoptr(GString) group_list = NULL;
 
     group_list = g_string_new ("");
     for (i = 0; i < size; i++)
@@ -132,7 +147,6 @@ setgroups (size_t size, const gid_t *list)
         g_string_append_printf (group_list, "%d", list[i]);
     }
     g_setenv ("LIGHTDM_TEST_GROUPS", group_list->str, TRUE);
-    g_string_free (group_list, TRUE);
 
     return 0;
 }
@@ -196,6 +210,9 @@ redirect_path (const gchar *path)
     if (g_str_has_prefix (path, "/tmp"))
         return g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "tmp", path + strlen ("/tmp"), NULL);
 
+    if (g_str_has_prefix (path, "/run"))
+        return g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "run", path + strlen ("/run"), NULL);
+
     if (g_str_has_prefix (path, "/etc/xdg"))
         return g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "etc", "xdg", path + strlen ("/etc/xdg"), NULL);
 
@@ -210,24 +227,23 @@ static int
 open_wrapper (const char *func, const char *pathname, int flags, mode_t mode)
 {
     int (*_open) (const char *pathname, int flags, mode_t mode);
-    gchar *new_path = NULL;
+    g_autofree gchar *new_path = NULL;
     int fd;
 
     _open = (int (*)(const char *pathname, int flags, mode_t mode)) dlsym (RTLD_NEXT, func);
 
-    if (strcmp (pathname, "/dev/console") == 0)
+    if (strcmp (pathname, "/dev/tty0") == 0)
     {
-        if (console_fd < 0)
+        if (tty_fd < 0)
         {
-            console_fd = _open ("/dev/null", flags, mode);
-            fcntl (console_fd, F_SETFD, FD_CLOEXEC);
+            tty_fd = _open ("/dev/null", flags, mode);
+            fcntl (tty_fd, F_SETFD, FD_CLOEXEC);
         }
-        return console_fd;
+        return tty_fd;
     }
 
     new_path = redirect_path (pathname);
     fd = _open (new_path, flags, mode);
-    g_free (new_path);
 
     return fd;
 }
@@ -264,14 +280,13 @@ FILE *
 fopen (const char *path, const char *mode)
 {
     FILE *(*_fopen) (const char *pathname, const char *mode);
-    gchar *new_path = NULL;
+    g_autofree gchar *new_path = NULL;
     FILE *result;
 
     _fopen = (FILE *(*)(const char *pathname, const char *mode)) dlsym (RTLD_NEXT, "fopen");
 
     new_path = redirect_path (path);
     result = _fopen (new_path, mode);
-    g_free (new_path);
 
     return result;
 }
@@ -280,14 +295,13 @@ int
 unlinkat (int dirfd, const char *pathname, int flags)
 {
     int (*_unlinkat) (int dirfd, const char *pathname, int flags);
-    gchar *new_path = NULL;
+    g_autofree gchar *new_path = NULL;
     int result;
 
     _unlinkat = (int (*)(int dirfd, const char *pathname, int flags)) dlsym (RTLD_NEXT, "unlinkat");
 
     new_path = redirect_path (pathname);
     result = _unlinkat (dirfd, new_path, flags);
-    g_free (new_path);
 
     return result;
 }
@@ -296,14 +310,13 @@ int
 creat (const char *pathname, mode_t mode)
 {
     int (*_creat) (const char *pathname, mode_t mode);
-    gchar *new_path = NULL;
+    g_autofree gchar *new_path = NULL;
     int result;
 
     _creat = (int (*)(const char *pathname, mode_t mode)) dlsym (RTLD_NEXT, "creat");
 
     new_path = redirect_path (pathname);
     result = _creat (new_path, mode);
-    g_free (new_path);
 
     return result;
 }
@@ -312,14 +325,13 @@ int
 creat64 (const char *pathname, mode_t mode)
 {
     int (*_creat64) (const char *pathname, mode_t mode);
-    gchar *new_path = NULL;
+    g_autofree gchar *new_path = NULL;
     int result;
 
     _creat64 = (int (*)(const char *pathname, mode_t mode)) dlsym (RTLD_NEXT, "creat64");
 
     new_path = redirect_path (pathname);
     result = _creat64 (new_path, mode);
-    g_free (new_path);
 
     return result;
 }
@@ -328,148 +340,113 @@ int
 access (const char *pathname, int mode)
 {
     int (*_access) (const char *pathname, int mode);
-    gchar *new_path = NULL;
-    int ret;
-
-    /* Look like systemd is always running */
-    if (strcmp (pathname, "/run/systemd/seats/") == 0)
-        return 1;
+    g_autofree gchar *new_path = NULL;
 
     _access = (int (*)(const char *pathname, int mode)) dlsym (RTLD_NEXT, "access");
 
-    new_path = redirect_path (pathname);
-    ret = _access (new_path, mode);
-    g_free (new_path);
+    if (strcmp (pathname, "/dev/tty0") == 0)
+        return F_OK;
+    if (strcmp (pathname, "/sys/class/tty/tty0/active") == 0)
+        return F_OK;
 
-    return ret;
+    new_path = redirect_path (pathname);
+    return _access (new_path, mode);
 }
 
 int
 stat (const char *path, struct stat *buf)
 {
     int (*_stat) (const char *path, struct stat *buf);
-    gchar *new_path = NULL;
-    int ret;
-  
+    g_autofree gchar *new_path = NULL;
+
     _stat = (int (*)(const char *path, struct stat *buf)) dlsym (RTLD_NEXT, "stat");
 
     new_path = redirect_path (path);
-    ret = _stat (new_path, buf);
-    g_free (new_path);
-
-    return ret;
+    return _stat (new_path, buf);
 }
 
 int
 stat64 (const char *path, struct stat64 *buf)
 {
     int (*_stat64) (const char *path, struct stat64 *buf);
-    gchar *new_path = NULL;
-    int ret;
+    g_autofree gchar *new_path = NULL;
 
     _stat64 = (int (*)(const char *path, struct stat64 *buf)) dlsym (RTLD_NEXT, "stat64");
 
     new_path = redirect_path (path);
-    ret = _stat64 (new_path, buf);
-    g_free (new_path);
-
-    return ret;
+    return _stat64 (new_path, buf);
 }
 
 int
 __xstat (int version, const char *path, struct stat *buf)
 {
     int (*___xstat) (int version, const char *path, struct stat *buf);
-    gchar *new_path = NULL;
-    int ret;
-  
+    g_autofree gchar *new_path = NULL;
+
     ___xstat = (int (*)(int version, const char *path, struct stat *buf)) dlsym (RTLD_NEXT, "__xstat");
 
     new_path = redirect_path (path);
-    ret = ___xstat (version, new_path, buf);
-    g_free (new_path);
-
-    return ret;
+    return ___xstat (version, new_path, buf);
 }
 
 int
 __xstat64 (int version, const char *path, struct stat64 *buf)
 {
     int (*___xstat64) (int version, const char *path, struct stat64 *buf);
-    gchar *new_path = NULL;
-    int ret;
-  
+    g_autofree gchar *new_path = NULL;
+
     ___xstat64 = (int (*)(int version, const char *path, struct stat64 *buf)) dlsym (RTLD_NEXT, "__xstat64");
 
     new_path = redirect_path (path);
-    ret = ___xstat64 (version, new_path, buf);
-    g_free (new_path);
-
-    return ret;
+    return ___xstat64 (version, new_path, buf);
 }
 
 int
 __fxstatat(int ver, int dirfd, const char *pathname, struct stat *buf, int flags)
 {
     int (*___fxstatat) (int ver, int dirfd, const char *pathname, struct stat *buf, int flags);
-    gchar *new_path = NULL;
-    int ret;
-  
+    g_autofree gchar *new_path = NULL;
+
     ___fxstatat = (int (*)(int ver, int dirfd, const char *pathname, struct stat *buf, int flags)) dlsym (RTLD_NEXT, "__fxstatat");
 
     new_path = redirect_path (pathname);
-    ret = ___fxstatat (ver, dirfd, new_path, buf, flags);
-    g_free (new_path);
-
-    return ret;
+    return ___fxstatat (ver, dirfd, new_path, buf, flags);
 }
 
 int
 __fxstatat64(int ver, int dirfd, const char *pathname, struct stat64 *buf, int flags)
 {
     int (*___fxstatat64) (int ver, int dirfd, const char *pathname, struct stat64 *buf, int flags);
-    gchar *new_path = NULL;
-    int ret;
-  
+    g_autofree gchar *new_path = NULL;
+
     ___fxstatat64 = (int (*)(int ver, int dirfd, const char *pathname, struct stat64 *buf, int flags)) dlsym (RTLD_NEXT, "__fxstatat64");
 
     new_path = redirect_path (pathname);
-    ret = ___fxstatat64 (ver, dirfd, new_path, buf, flags);
-    g_free (new_path);
-
-    return ret;
+    return ___fxstatat64 (ver, dirfd, new_path, buf, flags);
 }
 
 DIR *
 opendir (const char *name)
 {
     DIR *(*_opendir) (const char *name);
-    gchar *new_path = NULL;
-    DIR *result;
+    g_autofree gchar *new_path = NULL;
 
     _opendir = (DIR *(*)(const char *name)) dlsym (RTLD_NEXT, "opendir");
 
     new_path = redirect_path (name);
-    result = _opendir (new_path);
-    g_free (new_path);
-
-    return result; 
+    return _opendir (new_path);
 }
 
 int
 mkdir (const char *pathname, mode_t mode)
 {
     int (*_mkdir) (const char *pathname, mode_t mode);
-    gchar *new_path = NULL;
-    int result;
+    g_autofree gchar *new_path = NULL;
 
     _mkdir = (int (*)(const char *pathname, mode_t mode)) dlsym (RTLD_NEXT, "mkdir");
 
     new_path = redirect_path (pathname);
-    result = _mkdir (new_path, mode);
-    g_free (new_path);
-
-    return result;
+    return _mkdir (new_path, mode);
 }
 
 int
@@ -483,16 +460,12 @@ int
 chmod (const char *path, mode_t mode)
 {
     int (*_chmod) (const char *path, mode_t mode);
-    gchar *new_path = NULL;
-    int result;
+    g_autofree gchar *new_path = NULL;
 
     _chmod = (int (*)(const char *path, mode_t mode)) dlsym (RTLD_NEXT, "chmod");
 
     new_path = redirect_path (path);
-    result = _chmod (new_path, mode);
-    g_free (new_path);
-
-    return result;
+    return _chmod (new_path, mode);
 }
 
 int
@@ -501,9 +474,9 @@ ioctl (int d, unsigned long request, ...)
     int (*_ioctl) (int d, int request, ...);
 
     _ioctl = (int (*)(int d, int request, ...)) dlsym (RTLD_NEXT, "ioctl");
-    if (d > 0 && d == console_fd)
+    if (d > 0 && d == tty_fd)
     {
-        struct vt_stat *console_state;
+        struct vt_stat *vt_state;
         int vt;
         va_list ap;
 
@@ -511,9 +484,9 @@ ioctl (int d, unsigned long request, ...)
         {
         case VT_GETSTATE:
             va_start (ap, request);
-            console_state = va_arg (ap, struct vt_stat *);
+            vt_state = va_arg (ap, struct vt_stat *);
             va_end (ap);
-            console_state->v_active = active_vt;
+            vt_state->v_active = active_vt;
             break;
         case VT_ACTIVATE:
             va_start (ap, request);
@@ -543,12 +516,227 @@ ioctl (int d, unsigned long request, ...)
     }
 }
 
+static void
+add_port_redirect (int requested_port, int redirected_port)
+{
+    g_autoptr(GKeyFile) file = NULL;
+    g_autofree gchar *path = NULL;
+    g_autofree gchar *name = NULL;
+    g_autofree gchar *data = NULL;
+
+    file = g_key_file_new ();
+    path = g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), ".port-redirects", NULL);
+    g_key_file_load_from_file (file, path, G_KEY_FILE_NONE, NULL);
+
+    name = g_strdup_printf ("%d", requested_port);
+    g_key_file_set_integer (file, name, "redirected", redirected_port);
+  
+    data = g_key_file_to_data (file, NULL, NULL);
+    g_file_set_contents (path, data, -1, NULL);
+}
+
+static int
+find_port_redirect (int port)
+{
+    g_autoptr(GKeyFile) file = NULL;
+    g_autofree gchar *path = NULL;
+    g_autofree gchar *name = NULL;
+    int redirected_port;
+
+    file = g_key_file_new ();
+    path = g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), ".port-redirects", NULL);
+    g_key_file_load_from_file (file, path, G_KEY_FILE_NONE, NULL);
+
+    name = g_strdup_printf ("%d", port);
+    redirected_port = g_key_file_get_integer (file, name, "redirected", NULL);
+
+    return redirected_port;
+}
+
+int
+bind (int sockfd, const struct sockaddr *addr, socklen_t addrlen)
+{
+    int port = 0, redirected_port = 0;
+    const char *path;
+    int (*_bind) (int sockfd, const struct sockaddr *addr, socklen_t addrlen);
+    const struct sockaddr *modified_addr = addr;
+    struct sockaddr_in temp_addr_in;
+    struct sockaddr_in6 temp_addr_in6;
+    struct sockaddr_un temp_addr_un;
+    int retval;
+
+    _bind = (int (*)(int sockfd, const struct sockaddr *addr, socklen_t addrlen)) dlsym (RTLD_NEXT, "bind");
+
+    switch (addr->sa_family)
+    {
+    case AF_UNIX:
+        path = ((const struct sockaddr_un *) addr)->sun_path;
+        if (path[0] != '\0')
+        {
+            g_autofree gchar *new_path = redirect_path (path);
+            memcpy (&temp_addr_un, addr, sizeof (struct sockaddr_un));
+            strncpy (temp_addr_un.sun_path, new_path, sizeof (temp_addr_un.sun_path) - 1);
+            modified_addr = (struct sockaddr *) &temp_addr_un;
+        }
+        break;
+    case AF_INET:
+        port = ntohs (((const struct sockaddr_in *) addr)->sin_port);
+        redirected_port = find_port_redirect (port);
+        memcpy (&temp_addr_in, addr, sizeof (struct sockaddr_in));
+        modified_addr = (struct sockaddr *) &temp_addr_in;
+        if (redirected_port != 0)
+            temp_addr_in.sin_port = htons (redirected_port);
+        else
+            temp_addr_in.sin_port = 0;
+        break;
+    case AF_INET6:
+        port = ntohs (((const struct sockaddr_in6 *) addr)->sin6_port);
+        redirected_port = find_port_redirect (port);
+        memcpy (&temp_addr_in6, addr, sizeof (struct sockaddr_in6));
+        modified_addr = (struct sockaddr *) &temp_addr_in6;
+        if (redirected_port != 0)
+            temp_addr_in6.sin6_port = htons (redirected_port);
+        else
+            temp_addr_in6.sin6_port = 0;
+        break;
+    }
+
+    retval = _bind (sockfd, modified_addr, addrlen);
+
+    socklen_t temp_addr_len;
+    switch (addr->sa_family)
+    {
+    case AF_INET:
+        temp_addr_len = sizeof (temp_addr_in);
+        getsockname (sockfd, &temp_addr_in, &temp_addr_len);
+        if (redirected_port == 0)
+        {
+            redirected_port = ntohs (temp_addr_in.sin_port);
+            add_port_redirect (port, redirected_port);
+        }
+        break;
+    case AF_INET6:
+        temp_addr_len = sizeof (temp_addr_in6);
+        getsockname (sockfd, &temp_addr_in6, &temp_addr_len);
+        if (redirected_port == 0)
+        {
+            redirected_port = ntohs (temp_addr_in6.sin6_port);
+            add_port_redirect (port, redirected_port);
+        }
+        break;
+    }
+
+    return retval;
+}
+
+#include <ctype.h>
+
+int
+connect (int sockfd, const struct sockaddr *addr, socklen_t addrlen)
+{
+    int port, redirected_port;
+    const char *path;
+    const struct sockaddr *modified_addr = addr;
+    struct sockaddr_in temp_addr_in;
+    struct sockaddr_in6 temp_addr_in6;
+    struct sockaddr_un temp_addr_un;
+    int (*_connect) (int sockfd, const struct sockaddr *addr, socklen_t addrlen);
+
+    _connect = (int (*)(int sockfd, const struct sockaddr *addr, socklen_t addrlen)) dlsym (RTLD_NEXT, "connect");
+
+    switch (addr->sa_family)
+    {
+    case AF_UNIX:
+        path = ((const struct sockaddr_un *) addr)->sun_path;
+        if (path[0] != '\0') 
+        {
+            g_autofree gchar *new_path = redirect_path (path);
+            memcpy (&temp_addr_un, addr, addrlen);
+            strncpy (temp_addr_un.sun_path, new_path, sizeof (temp_addr_un.sun_path) - 1);
+            modified_addr = (struct sockaddr *) &temp_addr_un;
+        }
+        break;
+    case AF_INET:
+        port = ntohs (((const struct sockaddr_in *) addr)->sin_port);
+        redirected_port = find_port_redirect (port);
+        if (redirected_port != 0) 
+        {
+            memcpy (&temp_addr_in, addr, sizeof (struct sockaddr_in));
+            temp_addr_in.sin_port = htons (redirected_port);
+            modified_addr = (struct sockaddr *) &temp_addr_in;
+        }
+        break;
+    case AF_INET6:
+        port = ntohs (((const struct sockaddr_in6 *) addr)->sin6_port);
+        redirected_port = find_port_redirect (port);
+        if (redirected_port != 0) 
+        {
+            memcpy (&temp_addr_in6, addr, sizeof (struct sockaddr_in6));
+            temp_addr_in6.sin6_port = htons (redirected_port);
+            modified_addr = (struct sockaddr *) &temp_addr_in6;
+        }
+        break;
+    }
+
+    return _connect (sockfd, modified_addr, addrlen);
+}
+
+ssize_t
+sendto (int sockfd, const void *buf, size_t len, int flags, const struct sockaddr *dest_addr, socklen_t addrlen)
+{
+    int port, redirected_port;
+    const char *path;
+    const struct sockaddr *modified_addr = dest_addr;
+    struct sockaddr_in temp_addr_in;
+    struct sockaddr_in6 temp_addr_in6;
+    struct sockaddr_un temp_addr_un;  
+    ssize_t (*_sendto) (int sockfd, const void *buf, size_t len, int flags, const struct sockaddr *dest_addr, socklen_t addrlen);
+
+    _sendto = (ssize_t (*)(int sockfd, const void *buf, size_t len, int flags, const struct sockaddr *dest_addr, socklen_t addrlen)) dlsym (RTLD_NEXT, "sendto");
+
+    switch (dest_addr->sa_family)
+    {
+    case AF_UNIX:
+        path = ((const struct sockaddr_un *) dest_addr)->sun_path;
+        if (path[0] != '\0') 
+        {
+            g_autofree gchar *new_path = redirect_path (path);
+            memcpy (&temp_addr_un, dest_addr, sizeof (struct sockaddr_un));
+            strncpy (temp_addr_un.sun_path, new_path, sizeof (temp_addr_un.sun_path) - 1);
+            modified_addr = (struct sockaddr *) &temp_addr_un;
+        }
+        break;
+    case AF_INET:
+        port = ntohs (((const struct sockaddr_in *) dest_addr)->sin_port);
+        redirected_port = find_port_redirect (port);
+        if (redirected_port != 0) 
+        {
+            memcpy (&temp_addr_in, dest_addr, sizeof (struct sockaddr_in));
+            temp_addr_in.sin_port = htons (redirected_port);
+            modified_addr = (struct sockaddr *) &temp_addr_in;
+        }
+        break;
+    case AF_INET6:
+        port = ntohs (((const struct sockaddr_in6 *) dest_addr)->sin6_port);
+        redirected_port = find_port_redirect (port);
+        if (redirected_port != 0) 
+        {
+            memcpy (&temp_addr_in6, dest_addr, sizeof (struct sockaddr_in6));
+            temp_addr_in6.sin6_port = htons (redirected_port);
+            modified_addr = (struct sockaddr *) &temp_addr_in6;
+        }
+        break;
+    }
+
+    return _sendto (sockfd, buf, len, flags, modified_addr, addrlen);
+}
+
 int
 close (int fd)
 {
     int (*_close) (int fd);
 
-    if (fd > 0 && fd == console_fd)
+    if (fd > 0 && fd == tty_fd)
         return 0;
 
     _close = (int (*)(int fd)) dlsym (RTLD_NEXT, "close");
@@ -560,7 +748,7 @@ static void
 free_user (gpointer data)
 {
     struct passwd *entry = data;
-  
+
     g_free (entry->pw_name);
     g_free (entry->pw_passwd);
     g_free (entry->pw_gecos);
@@ -572,30 +760,29 @@ free_user (gpointer data)
 static void
 load_passwd_file (void)
 {
-    gchar *path, *data = NULL, **lines;
+    g_autofree gchar *path = NULL;
+    g_autofree gchar *data = NULL;
+    g_auto(GStrv) lines = NULL;
     gint i;
-    GError *error = NULL;
+    g_autoptr(GError) error = NULL;
 
     g_list_free_full (user_entries, free_user);
     user_entries = NULL;
     getpwent_link = NULL;
 
     path = g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "etc", "passwd", NULL);
-    g_file_get_contents (path, &data, NULL, &error);
-    g_free (path);
-    if (error)
+    if (!g_file_get_contents (path, &data, NULL, &error))
+    {
         g_warning ("Error loading passwd file: %s", error->message);
-    g_clear_error (&error);
-
-    if (!data)
         return;
+    }
 
     lines = g_strsplit (data, "\n", -1);
-    g_free (data);
 
     for (i = 0; lines[i]; i++)
     {
-        gchar *line, **fields;
+        const gchar *line;
+        g_auto(GStrv) fields = NULL;
 
         line = g_strstrip (lines[i]);
         fields = g_strsplit (line, ":", -1);
@@ -612,9 +799,7 @@ load_passwd_file (void)
             entry->pw_shell = g_strdup (fields[6]);
             user_entries = g_list_append (user_entries, entry);
         }
-        g_strfreev (fields);
     }
-    g_strfreev (lines);
 }
 
 struct passwd *
@@ -653,10 +838,7 @@ struct passwd *
 getpwnam (const char *name)
 {
     GList *link;
-  
-    if (name == NULL)
-        return NULL;
-  
+
     load_passwd_file ();
 
     for (link = user_entries; link; link = link->next)
@@ -694,7 +876,7 @@ static void
 free_group (gpointer data)
 {
     struct group *entry = data;
-  
+
     g_free (entry->gr_name);
     g_free (entry->gr_passwd);
     g_strfreev (entry->gr_mem);
@@ -704,29 +886,28 @@ free_group (gpointer data)
 static void
 load_group_file (void)
 {
-    gchar *path, *data = NULL, **lines;
+    g_autofree gchar *path = NULL;
+    g_autofree gchar *data = NULL;
+    g_auto(GStrv) lines = NULL;
     gint i;
-    GError *error = NULL;
+    g_autoptr(GError) error = NULL;
 
     g_list_free_full (group_entries, free_group);
     group_entries = NULL;
 
     path = g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), "etc", "group", NULL);
-    g_file_get_contents (path, &data, NULL, &error);
-    g_free (path);
-    if (error)
+    if (!g_file_get_contents (path, &data, NULL, &error))
+    {
         g_warning ("Error loading group file: %s", error->message);
-    g_clear_error (&error);
-
-    if (!data)
         return;
+    }
 
     lines = g_strsplit (data, "\n", -1);
-    g_free (data);
 
     for (i = 0; lines[i]; i++)
     {
-        gchar *line, **fields;
+        const gchar *line;
+        g_auto(GStrv) fields = NULL;
 
         line = g_strstrip (lines[i]);
         fields = g_strsplit (line, ":", -1);
@@ -740,9 +921,7 @@ load_group_file (void)
             entry->gr_mem = g_strsplit (fields[3], ",", -1);
             group_entries = g_list_append (group_entries, entry);
         }
-        g_strfreev (fields);
     }
-    g_strfreev (lines);
 }
 
 struct group *
@@ -788,9 +967,6 @@ pam_start (const char *service_name, const char *user, const struct pam_conv *co
 {
     pam_handle_t *handle;
 
-    if (service_name == NULL || conversation == NULL || pamh == NULL)
-        return PAM_SYSTEM_ERR;
-
     handle = *pamh = malloc (sizeof (pam_handle_t));
     if (handle == NULL)
         return PAM_BUF_ERR;
@@ -803,15 +979,14 @@ pam_start (const char *service_name, const char *user, const struct pam_conv *co
     connect_status ();
     if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
     {
-        GString *status;
+        g_autoptr(GString) status = NULL;
 
         status = g_string_new ("");
         g_string_append_printf (status, "%s START", handle->id);
         g_string_append_printf (status, " SERVICE=%s", service_name);
         if (user)
             g_string_append_printf (status, " USER=%s", user);
-        status_notify (status->str);
-        g_string_free (status, TRUE);
+        status_notify ("%s", status->str);
     }
 
     handle->service_name = strdup (service_name);
@@ -833,13 +1008,10 @@ pam_authenticate (pam_handle_t *pamh, int flags)
     struct passwd *entry;
     gboolean password_matches = FALSE;
 
-    if (pamh == NULL)
-        return PAM_SYSTEM_ERR;
-
     connect_status ();
     if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
     {
-        GString *status;
+        g_autoptr(GString) status = NULL;
 
         status = g_string_new ("");
         g_string_append_printf (status, "%s AUTHENTICATE", pamh->id);
@@ -848,10 +1020,9 @@ pam_authenticate (pam_handle_t *pamh, int flags)
         if (flags & PAM_DISALLOW_NULL_AUTHTOK)
             g_string_append (status, " DISALLOW_NULL_AUTHTOK");
 
-        status_notify (status->str);
-        g_string_free (status, TRUE);
+        status_notify ("%s", status->str);
     }
-  
+
     if (strcmp (pamh->service_name, "test-remote") == 0)
     {
         int result;
@@ -860,7 +1031,7 @@ pam_authenticate (pam_handle_t *pamh, int flags)
 
         msg = malloc (sizeof (struct pam_message *) * 1);
         msg[0] = malloc (sizeof (struct pam_message));
-        msg[0]->msg_style = PAM_PROMPT_ECHO_ON; 
+        msg[0]->msg_style = PAM_PROMPT_ECHO_ON;
         msg[0]->msg = "remote-login:";
         result = pamh->conversation.conv (1, (const struct pam_message **) msg, &resp, pamh->conversation.appdata_ptr);
         free (msg[0]);
@@ -923,7 +1094,7 @@ pam_authenticate (pam_handle_t *pamh, int flags)
 
         msg = malloc (sizeof (struct pam_message *) * 1);
         msg[0] = malloc (sizeof (struct pam_message));
-        msg[0]->msg_style = PAM_PROMPT_ECHO_ON; 
+        msg[0]->msg_style = PAM_PROMPT_ECHO_ON;
         msg[0]->msg = LOGIN_PROMPT;
         result = pamh->conversation.conv (1, (const struct pam_message **) msg, &resp, pamh->conversation.appdata_ptr);
         free (msg[0]);
@@ -938,7 +1109,7 @@ pam_authenticate (pam_handle_t *pamh, int flags)
             free (resp);
             return PAM_CONV_ERR;
         }
-      
+
         pamh->user = strdup (resp[0].resp);
         free (resp[0].resp);
         free (resp);
@@ -1034,6 +1205,8 @@ pam_authenticate (pam_handle_t *pamh, int flags)
             result = pamh->conversation.conv (1, (const struct pam_message **) msg, &resp, pamh->conversation.appdata_ptr);
             free (msg[0]);
             free (msg);
+            if (result != PAM_SUCCESS)
+                return result;
 
             if (resp == NULL)
                 return PAM_CONV_ERR;
@@ -1076,7 +1249,7 @@ static const char *
 get_env_value (const char *name_value, const char *name)
 {
     int j;
-  
+
     for (j = 0; name[j] && name_value[j] && name[j] == name_value[j]; j++);
     if (name[j] == '\0' && name_value[j] == '=')
         return &name_value[j + 1];
@@ -1090,9 +1263,6 @@ pam_putenv (pam_handle_t *pamh, const char *name_value)
     int i;
     gchar *name;
 
-    if (pamh == NULL || name_value == NULL)
-        return PAM_SYSTEM_ERR;
-
     name = strdup (name_value);
     for (i = 0; name[i]; i++)
         if (name[i] == '=')
@@ -1124,9 +1294,6 @@ pam_getenv (pam_handle_t *pamh, const char *name)
 {
     int i;
 
-    if (pamh == NULL || name == NULL)
-        return NULL;
-
     for (i = 0; pamh->envlist[i]; i++)
     {
         const char *value;
@@ -1141,16 +1308,13 @@ pam_getenv (pam_handle_t *pamh, const char *name)
 char **
 pam_getenvlist (pam_handle_t *pamh)
 {
-    if (pamh == NULL)
-        return NULL;
-
     return pamh->envlist;
 }
 
 int
 pam_set_item (pam_handle_t *pamh, int item_type, const void *item)
 {
-    if (pamh == NULL || item == NULL)
+    if (item == NULL)
         return PAM_SYSTEM_ERR;
 
     switch (item_type)
@@ -1169,15 +1333,15 @@ pam_set_item (pam_handle_t *pamh, int item_type, const void *item)
 int
 pam_get_item (const pam_handle_t *pamh, int item_type, const void **item)
 {
-    if (pamh == NULL || item == NULL)
+    if (item == NULL)
         return PAM_SYSTEM_ERR;
-  
+
     switch (item_type)
     {
     case PAM_SERVICE:
         *item = pamh->service_name;
         return PAM_SUCCESS;
-      
+
     case PAM_USER:
         *item = pamh->user;
         return PAM_SUCCESS;
@@ -1189,11 +1353,11 @@ pam_get_item (const pam_handle_t *pamh, int item_type, const void **item)
     case PAM_RUSER:
         *item = pamh->ruser;
         return PAM_SUCCESS;
-     
+
     case PAM_USER_PROMPT:
         *item = LOGIN_PROMPT;
         return PAM_SUCCESS;
-      
+
     case PAM_TTY:
         *item = pamh->tty;
         return PAM_SUCCESS;
@@ -1210,24 +1374,20 @@ pam_get_item (const pam_handle_t *pamh, int item_type, const void **item)
 int
 pam_open_session (pam_handle_t *pamh, int flags)
 {
-    GVariant *result;
-    GError *error = NULL;
-
-    if (pamh == NULL)
-        return PAM_SYSTEM_ERR;
+    g_autoptr(GVariant) result = NULL;
+    g_autoptr(GError) error = NULL;
 
     connect_status ();
     if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
     {
-        GString *status;
+        g_autoptr(GString) status = NULL;
 
         status = g_string_new ("");
         g_string_append_printf (status, "%s OPEN-SESSION", pamh->id);
         if (flags & PAM_SILENT)
             g_string_append (status, " SILENT");
 
-        status_notify (status->str);
-        g_string_free (status, TRUE);
+        status_notify ("%s", status->str);
     }
 
     if (strcmp (pamh->user, "session-error") == 0)
@@ -1254,18 +1414,15 @@ pam_open_session (pam_handle_t *pamh, int flags)
                                           &error);
     if (result)
     {
-        gchar *e;
+        g_autofree gchar *e = NULL;
         const gchar *id;
 
         g_variant_get (result, "(&so)", &id, NULL);
         e = g_strdup_printf ("XDG_SESSION_ID=%s", id);
         pam_putenv (pamh, e);
-        g_free (e);
-        g_variant_unref (result);
     }
     else
         g_printerr ("Failed to create logind session: %s\n", error->message);
-    g_clear_error (&error);
 
     return PAM_SUCCESS;
 }
@@ -1273,21 +1430,17 @@ pam_open_session (pam_handle_t *pamh, int flags)
 int
 pam_close_session (pam_handle_t *pamh, int flags)
 {
-    if (pamh == NULL)
-        return PAM_SYSTEM_ERR;
-
     connect_status ();
     if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
     {
-        GString *status;
+        g_autoptr(GString) status = NULL;
 
         status = g_string_new ("");
         g_string_append_printf (status, "%s CLOSE-SESSION", pamh->id);
         if (flags & PAM_SILENT)
             g_string_append (status, " SILENT");
 
-        status_notify (status->str);
-        g_string_free (status, TRUE);
+        status_notify ("%s", status->str);
     }
 
     return PAM_SUCCESS;
@@ -1296,13 +1449,10 @@ pam_close_session (pam_handle_t *pamh, int flags)
 int
 pam_acct_mgmt (pam_handle_t *pamh, int flags)
 {
-    if (pamh == NULL)
-        return PAM_SYSTEM_ERR;
-
     connect_status ();
     if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
     {
-        GString *status;
+        g_autoptr(GString) status = NULL;
 
         status = g_string_new ("");
         g_string_append_printf (status, "%s ACCT-MGMT", pamh->id);
@@ -1311,10 +1461,9 @@ pam_acct_mgmt (pam_handle_t *pamh, int flags)
         if (flags & PAM_DISALLOW_NULL_AUTHTOK)
             g_string_append (status, " DISALLOW_NULL_AUTHTOK");
 
-        status_notify (status->str);
-        g_string_free (status, TRUE);
+        status_notify ("%s", status->str);
     }
-  
+
     if (!pamh->user)
         return PAM_USER_UNKNOWN;
 
@@ -1336,13 +1485,10 @@ pam_chauthtok (pam_handle_t *pamh, int flags)
     struct pam_message **msg;
     struct pam_response *resp = NULL;
 
-    if (pamh == NULL)
-        return PAM_SYSTEM_ERR;
-
     connect_status ();
     if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
     {
-        GString *status;
+        g_autoptr(GString) status = NULL;
 
         status = g_string_new ("");
         g_string_append_printf (status, "%s CHAUTHTOK", pamh->id);
@@ -1351,8 +1497,7 @@ pam_chauthtok (pam_handle_t *pamh, int flags)
         if (flags & PAM_CHANGE_EXPIRED_AUTHTOK)
             g_string_append (status, " CHANGE_EXPIRED_AUTHTOK");
 
-        status_notify (status->str);
-        g_string_free (status, TRUE);
+        status_notify ("%s", status->str);
     }
 
     msg = malloc (sizeof (struct pam_message *) * 1);
@@ -1388,15 +1533,12 @@ pam_chauthtok (pam_handle_t *pamh, int flags)
 int
 pam_setcred (pam_handle_t *pamh, int flags)
 {
-    gchar *e;
-
-    if (pamh == NULL)
-        return PAM_SYSTEM_ERR;
+    g_autofree gchar *e = NULL;
 
     connect_status ();
     if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
     {
-        GString *status;
+        g_autoptr(GString) status = NULL;
 
         status = g_string_new ("");
         g_string_append_printf (status, "%s SETCRED", pamh->id);
@@ -1411,14 +1553,12 @@ pam_setcred (pam_handle_t *pamh, int flags)
         if (flags & PAM_REFRESH_CRED)
             g_string_append (status, " REFRESH_CRED");
 
-        status_notify (status->str);
-        g_string_free (status, TRUE);
+        status_notify ("%s", status->str);
     }
 
     /* Put the test directories into the path */
     e = g_strdup_printf ("PATH=%s/tests/src/.libs:%s/tests/src:%s/tests/src:%s/src:%s", BUILDDIR, BUILDDIR, SRCDIR, BUILDDIR, pam_getenv (pamh, "PATH"));
     pam_putenv (pamh, e);
-    g_free (e);
 
     if (strcmp (pamh->user, "cred-error") == 0)
         return PAM_CRED_ERR;
@@ -1460,18 +1600,14 @@ pam_setcred (pam_handle_t *pamh, int flags)
 int
 pam_end (pam_handle_t *pamh, int pam_status)
 {
-    if (pamh == NULL)
-        return PAM_SYSTEM_ERR;
-
     connect_status ();
     if (g_key_file_get_boolean (config, "test-pam", "log-events", NULL))
     {
-        GString *status;
+        g_autoptr(GString) status = NULL;
 
         status = g_string_new ("");
         g_string_append_printf (status, "%s END", pamh->id);
-        status_notify (status->str);
-        g_string_free (status, TRUE);
+        status_notify ("%s", status->str);
     }
 
     free (pamh->id);
@@ -1577,7 +1713,7 @@ pututxline (const struct utmpx *ut)
     connect_status ();
     if (g_key_file_get_boolean (config, "test-utmp-config", "check-events", NULL))
     {
-        GString *status;
+        g_autoptr(GString) status = NULL;
 
         status = g_string_new ("UTMP");
         switch (ut->ut_type)
@@ -1606,7 +1742,6 @@ pututxline (const struct utmpx *ut)
         if (ut->ut_host)
             g_string_append_printf (status, " HOST=%s", ut->ut_host);
         status_notify ("%s", status->str);
-        g_string_free (status, TRUE);
     }
 
     return (struct utmpx *)ut;
@@ -1623,7 +1758,7 @@ updwtmp (const char *wtmp_file, const struct utmp *ut)
     connect_status ();
     if (g_key_file_get_boolean (config, "test-utmp-config", "check-events", NULL))
     {
-        GString *status;
+        g_autoptr(GString) status = NULL;
 
         status = g_string_new ("WTMP");
         g_string_append_printf (status, " FILE=%s", wtmp_file);
@@ -1653,7 +1788,6 @@ updwtmp (const char *wtmp_file, const struct utmp *ut)
         if (ut->ut_host)
             g_string_append_printf (status, " HOST=%s", ut->ut_host);
         status_notify ("%s", status->str);
-        g_string_free (status, TRUE);
     }
 }
 
@@ -1668,9 +1802,7 @@ xcb_connection_t *
 xcb_connect_to_display_with_auth_info (const char *display, xcb_auth_info_t *auth, int *screen)
 {
     xcb_connection_t *c;
-    gchar *socket_path;
-    GError *error = NULL;
-  
+
     c = malloc (sizeof (xcb_connection_t));
     c->display = g_strdup (display);
     c->error = 0;
@@ -1682,32 +1814,32 @@ xcb_connect_to_display_with_auth_info (const char *display, xcb_auth_info_t *aut
 
     if (c->error == 0)
     {
+        g_autoptr(GError) error = NULL;
+
         c->socket = g_socket_new (G_SOCKET_FAMILY_UNIX, G_SOCKET_TYPE_STREAM, G_SOCKET_PROTOCOL_DEFAULT, &error);
-        if (error)
-            g_printerr ("%s\n", error->message);
-        g_clear_error (&error);
         if (c->socket == NULL)
+        {
+            g_printerr ("%s\n", error->message);
             c->error = XCB_CONN_ERROR;
+        }
     }
 
     if (c->error == 0)
     {
-        gchar *d;
-        GSocketAddress *address;
+        g_autofree gchar *d = NULL;
+        g_autofree gchar *socket_path = NULL;
+        g_autoptr(GSocketAddress) address = NULL;
+        g_autoptr(GError) error = NULL;
 
         /* Skip the hostname, we'll assume it's localhost */
         d = g_strdup_printf (".x%s", strchr (display, ':'));
-
         socket_path = g_build_filename (g_getenv ("LIGHTDM_TEST_ROOT"), d, NULL);
-        g_free (d);
         address = g_unix_socket_address_new (socket_path);
         if (!g_socket_connect (c->socket, address, NULL, &error))
-            c->error = XCB_CONN_ERROR;
-        g_object_unref (address);
-        if (error)
+        {
             g_printerr ("Failed to connect to X socket %s: %s\n", socket_path, error->message);
-        g_free (socket_path);
-        g_clear_error (&error);
+            c->error = XCB_CONN_ERROR;
+        }
     }
 
     // FIXME: Send auth info
@@ -1738,3 +1870,46 @@ xcb_disconnect (xcb_connection_t *c)
         g_object_unref (c->socket);
     free (c);
 }
+
+#if HAVE_LIBAUDIT
+int
+audit_open (void)
+{
+    connect_status ();  
+    if (g_key_file_get_boolean (config, "test-audit-config", "check-events", NULL))
+        status_notify ("AUDIT OPEN");
+
+    return dup (STDOUT_FILENO);
+}
+
+int
+audit_log_acct_message (int audit_fd, int type, const char *pgname,
+                        const char *op, const char *name, unsigned int id,
+                        const char *host, const char *addr, const char *tty, int result)
+{
+    g_autofree gchar *type_string = NULL;
+
+    connect_status ();  
+    if (!g_key_file_get_boolean (config, "test-audit-config", "check-events", NULL))
+        return 1;
+
+    switch (type)
+    {
+    case AUDIT_USER_LOGIN:
+        type_string = g_strdup ("USER_LOGIN");
+        break;      
+    case AUDIT_USER_LOGOUT:
+        type_string = g_strdup ("USER_LOGOUT");
+        break;
+    default:
+        type_string = g_strdup_printf ("%d", type);
+        break;
+    }
+  
+    status_notify ("AUDIT LOG-ACCT TYPE=%s PGNAME=%s OP=%s NAME=%s ID=%u HOST=%s ADDR=%s TTY=%s RESULT=%d",
+                   type_string, pgname ? pgname : "", op ? op : "", name ? name : "", id, host ? host : "", addr ? addr : "", tty ? tty : "", result);
+
+    return 1;
+}
+
+#endif
