@@ -36,18 +36,19 @@ struct XDMCPClientPrivate
     gchar *host;
     gint port;
     GSocket *socket;
-    guint query_timer;
+    gchar *authentication_names;
     gchar *authorization_name;
     gint authorization_data_length;
     guint8 *authorization_data;
 };
 
 enum {
-    XDMCP_CLIENT_QUERY,  
     XDMCP_CLIENT_WILLING,
+    XDMCP_CLIENT_UNWILLING,  
     XDMCP_CLIENT_ACCEPT,
     XDMCP_CLIENT_DECLINE,
     XDMCP_CLIENT_FAILED,
+    XDMCP_CLIENT_ALIVE,  
     XDMCP_CLIENT_LAST_SIGNAL
 };
 static guint xdmcp_client_signals[XDMCP_CLIENT_LAST_SIGNAL] = { 0 };
@@ -56,14 +57,13 @@ static void
 xdmcp_write (XDMCPClient *client, const guint8 *buffer, gssize buffer_length)
 {
     gssize n_written;
-    GError *error = NULL;
+    g_autoptr(GError) error = NULL;
 
     n_written = g_socket_send (client->priv->socket, (const gchar *) buffer, buffer_length, NULL, &error);
     if (n_written < 0)
         g_warning ("Failed to send XDMCP request: %s", error->message);
     else if (n_written != buffer_length)
         g_warning ("Partial write for XDMCP request, wrote %zi, expected %zi", n_written, buffer_length);
-    g_clear_error (&error);
 }
 
 static void
@@ -73,16 +73,6 @@ decode_willing (XDMCPClient *client, const guint8 *buffer, gssize buffer_length)
     gsize offset = 0;
     guint16 length;
 
-    if (client->priv->query_timer == 0)
-    {
-        g_debug ("Ignoring XDMCP unrequested/duplicate Willing");
-        return;
-    }
-
-    /* Stop sending queries */
-    g_source_remove (client->priv->query_timer);
-    client->priv->query_timer = 0;
-  
     message = g_malloc0 (sizeof (XDMCPWilling));
 
     length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
@@ -93,13 +83,34 @@ decode_willing (XDMCPClient *client, const guint8 *buffer, gssize buffer_length)
     message->status = read_string (buffer, buffer_length, length, &offset);
 
     g_signal_emit (client, xdmcp_client_signals[XDMCP_CLIENT_WILLING], 0, message);
-  
+
     g_free (message->authentication_name);
     g_free (message->hostname);
     g_free (message->status);
     g_free (message);
 }
 
+static void
+decode_unwilling (XDMCPClient *client, const guint8 *buffer, gssize buffer_length)
+{
+    XDMCPUnwilling *message;
+    gsize offset = 0;
+    guint16 length;
+
+    message = g_malloc0 (sizeof (XDMCPUnwilling));
+
+    length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
+    message->hostname = read_string (buffer, buffer_length, length, &offset);
+    length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
+    message->status = read_string (buffer, buffer_length, length, &offset);
+
+    g_signal_emit (client, xdmcp_client_signals[XDMCP_CLIENT_UNWILLING], 0, message);
+
+    g_free (message->hostname);
+    g_free (message->status);
+    g_free (message);
+}
+
 static void
 decode_accept (XDMCPClient *client, const guint8 *buffer, gssize buffer_length)
 {
@@ -112,8 +123,8 @@ decode_accept (XDMCPClient *client, const guint8 *buffer, gssize buffer_length)
     message->session_id = read_card32 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
     length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
     message->authentication_name = read_string (buffer, buffer_length, length, &offset);
-    length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
-    read_string8 (buffer, buffer_length, length, &offset);
+    message->authentication_data_length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
+    message->authentication_data = read_string8 (buffer, buffer_length, message->authentication_data_length, &offset);
     length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
     message->authorization_name = read_string (buffer, buffer_length, length, &offset);
     message->authorization_data_length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
@@ -122,6 +133,7 @@ decode_accept (XDMCPClient *client, const guint8 *buffer, gssize buffer_length)
     g_signal_emit (client, xdmcp_client_signals[XDMCP_CLIENT_ACCEPT], 0, message);
 
     g_free (message->authentication_name);
+    g_free (message->authentication_data);  
     g_free (message->authorization_name);
     g_free (message->authorization_data);
     g_free (message);
@@ -133,20 +145,21 @@ decode_decline (XDMCPClient *client, const guint8 *buffer, gssize buffer_length)
     XDMCPDecline *message;
     gsize offset = 0;
     guint16 length;
-  
+
     message = g_malloc0 (sizeof (XDMCPDecline));
 
     length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
     message->status = read_string (buffer, buffer_length, length, &offset);
     length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
     message->authentication_name = read_string (buffer, buffer_length, length, &offset);
-    length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
-    read_string8 (buffer, buffer_length, length, &offset);
-  
+    message->authentication_data_length = read_card16 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
+    message->authentication_data = read_string8 (buffer, buffer_length, message->authentication_data_length, &offset);
+
     g_signal_emit (client, xdmcp_client_signals[XDMCP_CLIENT_DECLINE], 0, message);
 
     g_free (message->status);
     g_free (message->authentication_name);
+    g_free (message->authentication_data);  
     g_free (message);
 }
 
@@ -169,6 +182,22 @@ decode_failed (XDMCPClient *client, const guint8 *buffer, gssize buffer_length)
     g_free (message);
 }
 
+static void
+decode_alive (XDMCPClient *client, const guint8 *buffer, gssize buffer_length)
+{
+    XDMCPAlive *message;
+    gsize offset = 0;
+
+    message = g_malloc0 (sizeof (XDMCPAlive));
+
+    message->session_running = read_card8 (buffer, buffer_length, &offset) != 0 ? TRUE : FALSE;
+    message->session_id = read_card32 (buffer, buffer_length, X_BYTE_ORDER_MSB, &offset);
+
+    g_signal_emit (client, xdmcp_client_signals[XDMCP_CLIENT_ALIVE], 0, message);
+
+    g_free (message);
+}
+
 static gboolean
 xdmcp_data_cb (GIOChannel *channel, GIOCondition condition, gpointer data)
 {
@@ -209,6 +238,10 @@ xdmcp_data_cb (GIOChannel *channel, GIOCondition condition, gpointer data)
             decode_willing (client, buffer + offset, n_read - offset);
             break;
 
+        case XDMCP_Unwilling:
+            decode_unwilling (client, buffer + offset, n_read - offset);
+            break;
+
         case XDMCP_Accept:
             decode_accept (client, buffer + offset, n_read - offset);
             break;
@@ -221,6 +254,10 @@ xdmcp_data_cb (GIOChannel *channel, GIOCondition condition, gpointer data)
             decode_failed (client, buffer + offset, n_read - offset);
             break;
 
+        case XDMCP_Alive:
+            decode_alive (client, buffer + offset, n_read - offset);
+            break;
+
         default:
             g_debug ("Ignoring unknown XDMCP opcode %d", opcode);
             break;
@@ -230,15 +267,6 @@ xdmcp_data_cb (GIOChannel *channel, GIOCondition condition, gpointer data)
     return TRUE;
 }
 
-static gboolean
-xdmcp_query_cb (gpointer data)
-{
-    XDMCPClient *client = data;
-    g_signal_emit (client, xdmcp_client_signals[XDMCP_CLIENT_QUERY], 0);
-    xdmcp_client_send_query (client);
-    return TRUE;
-}
-
 XDMCPClient *
 xdmcp_client_new (void)
 {
@@ -263,8 +291,10 @@ xdmcp_client_start (XDMCPClient *client)
 {
     GSocketConnectable *address;
     GSocketAddressEnumerator *enumerator;
-    gboolean result;
-    GError *error = NULL;
+    g_autoptr(GError) error = NULL;
+
+    if (client->priv->socket)
+        return TRUE;
 
     client->priv->socket = g_socket_new (G_SOCKET_FAMILY_IPV4, G_SOCKET_TYPE_DATAGRAM, G_SOCKET_PROTOCOL_UDP, &error);
     if (error)
@@ -274,45 +304,34 @@ xdmcp_client_start (XDMCPClient *client)
 
     address = g_network_address_new (client->priv->host, client->priv->port);
     enumerator = g_socket_connectable_enumerate (address);
-    result = FALSE;
-    while (TRUE) 
+    while (TRUE)
     {
-        GSocketAddress *socket_address;
-        GError *e = NULL;
+        g_autoptr(GSocketAddress) socket_address = NULL;
+        g_autoptr(GError) e = NULL;
 
         socket_address = g_socket_address_enumerator_next (enumerator, NULL, &e);
         if (e)
             g_warning ("Failed to get socket address: %s", e->message);
-        g_clear_error (&e);
         if (!socket_address)
-            break;
+            return FALSE;
 
-        result = g_socket_connect (client->priv->socket, socket_address, NULL, error ? NULL : &error);
-        g_object_unref (socket_address);
-        if (result)
+        if (!g_socket_connect (client->priv->socket, socket_address, NULL, &e))
         {
-            g_clear_error (&error);
-            break;
+            g_warning ("Unable to connect XDMCP socket: %s", error->message);
+            continue;
         }
+
+        g_io_add_watch (g_io_channel_unix_new (g_socket_get_fd (client->priv->socket)), G_IO_IN, xdmcp_data_cb, client);
+
+        return TRUE;
     }
-    if (error)
-        g_warning ("Unable to connect XDMCP socket: %s", error->message);
-    if (!result)
-        return FALSE;
-  
-    g_io_add_watch (g_io_channel_unix_new (g_socket_get_fd (client->priv->socket)), G_IO_IN, xdmcp_data_cb, client);
-
-    client->priv->query_timer = g_timeout_add (2000, xdmcp_query_cb, client);
-    xdmcp_query_cb (client); 
-
-    return TRUE;
 }
 
 GInetAddress *
 xdmcp_client_get_local_address (XDMCPClient *client)
 {
     GSocketAddress *socket_address;
-  
+
     if (!client->priv->socket)
         return NULL;
 
@@ -327,18 +346,48 @@ xdmcp_client_init (XDMCPClient *client)
     client->priv->port = XDMCP_PORT;
 }
 
-void
-xdmcp_client_send_query (XDMCPClient *client)
+static void
+send_query (XDMCPClient *client, guint16 opcode, gchar **authentication_names)
 {
     guint8 buffer[MAXIMUM_REQUEST_LENGTH];
-    gsize offset = 0;
+    gsize length, offset = 0, n_names = 0;
+    gchar **name;
+
+    length = 1;
+    for (name = authentication_names; authentication_names && *name; name++)
+    {
+        length += 2 + strlen (*name);
+        n_names++;
+    }
 
     write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, XDMCP_VERSION, &offset);
-    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, XDMCP_Query, &offset);
-    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, 1, &offset);
-    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &offset);
+    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, opcode, &offset);
+    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, length, &offset);
+    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, n_names, &offset);
+    for (name = authentication_names; authentication_names && *name; name++)
+    {
+        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, strlen (*name), &offset);
+        write_string (buffer, MAXIMUM_REQUEST_LENGTH, *name, &offset);      
+    }
+    xdmcp_write (client, buffer, offset);
+}
 
-    xdmcp_write (client, buffer, offset);   
+void
+xdmcp_client_send_query (XDMCPClient *client, gchar **authentication_names)
+{
+    send_query (client, XDMCP_Query, authentication_names);
+}
+
+void
+xdmcp_client_send_broadcast_query (XDMCPClient *client, gchar **authentication_names)
+{
+    send_query (client, XDMCP_BroadcastQuery, authentication_names);
+}
+
+void
+xdmcp_client_send_indirect_query (XDMCPClient *client, gchar **authentication_names)
+{
+    send_query (client, XDMCP_IndirectQuery, authentication_names);
 }
 
 void
@@ -403,7 +452,7 @@ xdmcp_client_send_request (XDMCPClient *client,
 }
 
 void
-xdmcp_client_send_manage (XDMCPClient *client, guint32 session_id, guint16 display_number, gchar *display_class)
+xdmcp_client_send_manage (XDMCPClient *client, guint32 session_id, guint16 display_number, const gchar *display_class)
 {
     guint8 buffer[MAXIMUM_REQUEST_LENGTH];
     gsize offset = 0;
@@ -417,18 +466,33 @@ xdmcp_client_send_manage (XDMCPClient *client, guint32 session_id, guint16 displ
     write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, strlen (display_class), &offset);
     write_string (buffer, MAXIMUM_REQUEST_LENGTH, display_class, &offset);
 
-    xdmcp_write (client, buffer, offset); 
+    xdmcp_write (client, buffer, offset);
+}
+
+void
+xdmcp_client_send_keep_alive (XDMCPClient *client, guint16 display_number, guint32 session_id)
+{
+    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
+    gsize offset = 0;
+
+    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, XDMCP_VERSION, &offset);
+    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, XDMCP_KeepAlive, &offset);
+    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, 6, &offset);
+
+    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, display_number, &offset);
+    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, X_BYTE_ORDER_MSB, session_id, &offset);
+
+    xdmcp_write (client, buffer, offset);
 }
 
 static void
 xdmcp_client_finalize (GObject *object)
 {
     XDMCPClient *client = (XDMCPClient *) object;
-    g_free (client->priv->host);
-    if (client->priv->socket)
-        g_object_unref (client->priv->socket);
-    g_free (client->priv->authorization_name);
-    g_free (client->priv->authorization_data);
+    g_clear_pointer (&client->priv->host, g_free);
+    g_clear_object (&client->priv->socket);
+    g_clear_pointer (&client->priv->authorization_name, g_free);
+    g_clear_pointer (&client->priv->authorization_data, g_free);
 }
 
 static void
@@ -437,44 +501,52 @@ xdmcp_client_class_init (XDMCPClientClass *klass)
     GObjectClass *object_class = G_OBJECT_CLASS (klass);
     object_class->finalize = xdmcp_client_finalize;
     g_type_class_add_private (klass, sizeof (XDMCPClientPrivate));
-    xdmcp_client_signals[XDMCP_CLIENT_QUERY] =
-        g_signal_new ("query",
-                      G_TYPE_FROM_CLASS (klass),
-                      G_SIGNAL_RUN_LAST,
-                      G_STRUCT_OFFSET (XDMCPClientClass, query),
-                      NULL, NULL,
-                      g_cclosure_marshal_VOID__VOID,
-                      G_TYPE_NONE, 0);
     xdmcp_client_signals[XDMCP_CLIENT_WILLING] =
-        g_signal_new ("willing",
+        g_signal_new (XDMCP_CLIENT_SIGNAL_WILLING,
                       G_TYPE_FROM_CLASS (klass),
                       G_SIGNAL_RUN_LAST,
                       G_STRUCT_OFFSET (XDMCPClientClass, willing),
                       NULL, NULL,
-                      g_cclosure_marshal_VOID__POINTER,
+                      NULL,
+                      G_TYPE_NONE, 1, G_TYPE_POINTER);
+    xdmcp_client_signals[XDMCP_CLIENT_UNWILLING] =
+        g_signal_new (XDMCP_CLIENT_SIGNAL_UNWILLING,
+                      G_TYPE_FROM_CLASS (klass),
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET (XDMCPClientClass, unwilling),
+                      NULL, NULL,
+                      NULL,
                       G_TYPE_NONE, 1, G_TYPE_POINTER);
     xdmcp_client_signals[XDMCP_CLIENT_ACCEPT] =
-        g_signal_new ("accept",
+        g_signal_new (XDMCP_CLIENT_SIGNAL_ACCEPT,
                       G_TYPE_FROM_CLASS (klass),
                       G_SIGNAL_RUN_LAST,
                       G_STRUCT_OFFSET (XDMCPClientClass, accept),
                       NULL, NULL,
-                      g_cclosure_marshal_VOID__POINTER,
+                      NULL,
                       G_TYPE_NONE, 1, G_TYPE_POINTER);
     xdmcp_client_signals[XDMCP_CLIENT_DECLINE] =
-        g_signal_new ("decline",
+        g_signal_new (XDMCP_CLIENT_SIGNAL_DECLINE,
                       G_TYPE_FROM_CLASS (klass),
                       G_SIGNAL_RUN_LAST,
                       G_STRUCT_OFFSET (XDMCPClientClass, decline),
                       NULL, NULL,
-                      g_cclosure_marshal_VOID__POINTER,
+                      NULL,
                       G_TYPE_NONE, 1, G_TYPE_POINTER);
     xdmcp_client_signals[XDMCP_CLIENT_FAILED] =
-        g_signal_new ("failed",
+        g_signal_new (XDMCP_CLIENT_SIGNAL_FAILED,
                       G_TYPE_FROM_CLASS (klass),
                       G_SIGNAL_RUN_LAST,
                       G_STRUCT_OFFSET (XDMCPClientClass, failed),
                       NULL, NULL,
-                      g_cclosure_marshal_VOID__POINTER,
+                      NULL,
+                      G_TYPE_NONE, 1, G_TYPE_POINTER);
+    xdmcp_client_signals[XDMCP_CLIENT_ALIVE] =
+        g_signal_new (XDMCP_CLIENT_SIGNAL_ALIVE,
+                      G_TYPE_FROM_CLASS (klass),
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET (XDMCPClientClass, alive),
+                      NULL, NULL,
+                      NULL,
                       G_TYPE_NONE, 1, G_TYPE_POINTER);
 }
